#pragma once
#include "MatBaza.h"
#include <cstdint>

/** Tip osnovnog kanala boje.
 *
 *  Ova klasa jednostavno predstavlja kanal kao realan broj i nudi sve očekivane operacije nad njemu koje su potrebne kod miješanja boja i različitih efekata sjenčanja. Svi računi se odvijaju barem u jednostrukoj preciznosti, a konverzija u standardni diskretni raspon će se dogoditi jedanput i to što kasnije u grafičkom cjevovodu. Algebra nad kanalima nije podložna provjerama prelijevanja, već se to ostavlja operatorima pretvorbe koji se pozivaju kada treba staviti sliku na ekran.
 */
class Kanal8
{
 public:
  //uint8_t broj;
  /** Vrijednost kanala.
   */
  Realan broj;
  /** Konstruktor kanala s realnim brojem.
   *  \param[in] broj Broj iz \f$[0,1]\f$.
   */
  __device__ __host__ Kanal8(Realan broj);
  /** Konstruktor kanala sa cijelim brojem.
   *  \param[in] broj Cijeli broj iz \f$[0,255]\f$.
   */
  __device__ __host__ Kanal8(unsigned int broj = 0);
  __device__ __host__ ~Kanal8() = default;
  
  __device__ __host__ Kanal8(const Kanal8 &) = default;
  __device__ __host__ Kanal8 & operator=(const Kanal8 &) = default;
  __device__ __host__ Kanal8(Kanal8 &&) = default;
  __device__ __host__ Kanal8 & operator=(Kanal8 &&) = default;

  /** Sumira dva kanala.
   *  \param[in] drugi Desni operand.
   *  \return Suma kanala.
   */
  __device__ __host__ Kanal8 operator+(const Kanal8 &drugi);
  /** Računa razliku dvaju kanala.
   *  \param[in] drugi Desni operand.
   *  \return Razlika kanala.
   */
  __device__ __host__ Kanal8 operator-(const Kanal8 &drugi);
  /** Računa umnožak dvaju kanala.
   *  \param[in] drugi Desni operand.
   *  \return Umnožak kanala.
   */
  __device__ __host__ Kanal8 operator*(const Kanal8 &drugi);
  /** Računa kvocijent dvaju kanala.
   *  \param[in] drugi Desni operand.
   *  \return Kvocijent kanala.
   */
  __device__ __host__ Kanal8 operator/(const Kanal8 &drugi);
  /** Računa umnožak kanala i proizvoljnog realnog broja.
   *  \param[in] broj Realan broj.
   *  \param[in] kan  Desni operand (kanal).
   *  \return Traženi umnožak.
   */
  friend __device__ __host__ Kanal8 operator*(Realan broj, Kanal8 kan);
  __device__ __host__ void operator+=(const Kanal8 &);
  __device__ __host__ void operator*=(const Kanal8 &);

  /** Pretvara kanal u realan broj.
   *
   *  Rezultat je iz raspona \f$[0,1]\f$, s tim da se ovdje *clampa* tj. ograničava na maksimum/minimum ako je rezultat prošlih operacija izašao van tog raspona.
   */
  __device__ __host__ operator Realan();
  /** Pretvara kanal u cijeli broj.
   *
   *  Rezultat je cijeli broj iz raspona \f$[0,255]\f$, s *clampanjem* po potrebi.
   */
  __device__ __host__ operator uint8_t();
 protected:
 private:
  //double limit;
};
