#include "KvadarGPU.h"
#include <cfloat>

__device__ __host__ KvadarGPU::KvadarGPU(Realan a, Realan b, Realan c, RGB boja)
{
  this->a = a;
  this->b = b;
  this->c = c;

  Vektor4 v1 = Vektor4(0, 0, 0, 1);
  Vektor4 v2 = Vektor4(a, 0, 0, 1);
  Vektor4 v3 = Vektor4(a, 0, c, 1);
  Vektor4 v4 = Vektor4(0, 0, c, 1);
  Vektor4 v5 = Vektor4(0, b, 0, 1);
  Vektor4 v6 = Vektor4(a, b, 0, 1);
  Vektor4 v7 = Vektor4(a, b, c, 1);
  Vektor4 v8 = Vektor4(0, b, c, 1);

  t[0] = TrokutGPU(v1, v2, v3, boja);
  t[1] = TrokutGPU(v1, v3, v4, boja);

  t[2] = TrokutGPU(v2, v1, v5, boja);
  t[3] = TrokutGPU(v5, v6, v2, boja);

  t[4] = TrokutGPU(v7, v3, v2, boja);
  t[5] = TrokutGPU(v6, v7, v2, boja);

  t[6] = TrokutGPU(v8, v4, v3, boja);
  t[7] = TrokutGPU(v3, v7, v8, boja);

  t[8] = TrokutGPU(v7, v6, v5, boja);
  t[9] = TrokutGPU(v5, v8, v7, boja);

  t[10] = TrokutGPU(v5, v1, v4, boja);
  t[11] = TrokutGPU(v4, v8, v5, boja);
}

__device__ __host__ void KvadarGPU::Transformiraj()
{
  for(int i = 0; i < 12; ++i)
  {
    t[i].transformacija = transformacija;
    t[i].Transformiraj();
  }
}

__device__ __host__ bool KvadarGPU::PresijeciZraku(const Zraka &zraka, Realan &t0, int &presjecenTrokut, Vektor3 &sjeciste)
{
  Realan t1 = BESKONACNO;
  Vektor3 sjeciste2, s;
  for(int i = 0; i < 12; ++i)
  {
    if(t[i].PresijeciZraku(zraka, t0, sjeciste2))
    {
      if(t0 < t1)
      {
        t1 = t0;
        presjecenTrokut = i;
	s = sjeciste2;
      }
    }
  }

  if(presjecenTrokut == -1)//if(t1 == DBL_MAX)
    return false;

  t0 = t1;
  sjeciste = s;
  return true;
}

__device__ __host__ RGB KvadarGPU::Sjencaj(int presjecenTrokut, const Vektor3 &sjeciste, const Svjetlo svjetla[], size_t n)
{
  return t[presjecenTrokut].Sjencaj(sjeciste, svjetla, n);
}
