#pragma once
#include <cfloat>
#if defined(__HIP__)
#include <hip/hip_runtime.h>
#else
#include <cmath>
#endif

/** \file MatBaza.h
 *  Zaglavlje s osnovnim konstantama i definicijama.
 */

typedef float Realan;
#if defined(__HIP__)
typedef float3 Real3;
typedef float4 Real4;
#endif

/** Konstanta koja predstavlja maksimalan koristan float broj.
 *
 *  Korišteno primarno radi pronalaženja minimalnih sjecišta zraka.
 */
const float BESKONACNO = FLT_MAX;
//const float M_PI = 3.1415926535897932385f;

/** Klasa koja nudi neke osnovne matematičke funkcije koje nisu dostupne u standardnim/HIP matematičkim bibliotekama.
 */
class MatBaza
{
 public:
  /** Pretvara dani broj stupnjeva (°) u radijane (rad).
   * \param[in] stupnjeva Broj stupnjeva.
   * \return Ekvivalentan broj radijana.
   */
  static Realan Radijana(Realan stupnjeva)
  {
    return (stupnjeva / 180 * M_PI);
  }

  /** Signum (sign) funkcija za realne brojeve.
   * \param[in] broj Realan broj.
   * \return sign(broj)
   */
  static Realan sign(Realan broj)
  {
    if (broj > 0)
      return 1;
    else if(broj == 0)
      return 0;
    else
      return -1;
  }
};
