#include "Matrica3x3.h"
#include <iostream>
#include <cstring>
//#include <cmath>

__device__ __host__ Matrica3x3::Matrica3x3()
{
  memset(m, 0, 9 * sizeof(Realan));
}

__device__ __host__ Matrica3x3::Matrica3x3(Realan m11, Realan m12, Realan m13, Realan m21, Realan m22, Realan m23, Realan m31, Realan m32, Realan m33)
{
  m[2][2] = m33;
  m[2][1] = m32;
  m[2][0] = m31;
  m[1][2] = m23;
  m[1][1] = m22;
  m[1][0] = m21;
  m[0][2] = m13;
  m[0][1] = m12;
  m[0][0] = m11;
}

__device__ __host__ Matrica3x3::Matrica3x3(const Realan elementi[3][3])
{
  memcpy (m, elementi, 9 * sizeof(Realan));
}

__device__ __host__ Matrica3x3::Matrica3x3(const Vektor3 stupci[3])
{
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
      {
	m[i][j] = stupci[j][i];
      }
}

/*
__device__ __host__ Matrica3x3::Matrica3x3 (const Matrica3x3 &mat)
{
  memcpy (m, mat.m, 9 * sizeof(Realan));
}

Matrica3x3 & Matrica3x3::operator=(const Matrica3x3 &mat)
{
  memcpy (m, mat.m, 9 * sizeof(Realan));
  return *this;
}
*/
/*
Matrica4x4::Matrica4x4 (const Matrica4x4 &&mat)
{
  memcpy (m, mat.m, 16 * sizeof(double));
  memset (mat.m, 0, 16 * sizeof(double));
}
*/
/*
Matrica4x4 & Matrica4x4::operator=(const Matrica4x4 &&mat)
{
  memcpy (m, mat.m, 16 * sizeof(double));
  memset (mat.m, 0, 16 * sizeof(double));
  return *this;
}
*/

__device__ __host__ Matrica3x3 Matrica3x3::operator+(const Matrica3x3 &mat) const
{
  Realan n[3][3];
  for (int i = 0; i < 3; ++i)
    for (int j = 0; j < 3; ++j)
      n[i][j] = m[i][j] + mat.Element(i, j);

  return Matrica3x3(n);
}

__device__ __host__ Matrica3x3 Matrica3x3::operator-(const Matrica3x3 &mat) const
{
  Realan n[3][3];
  for (int i = 0; i < 3; ++i)
    for (int j = 0; j < 3; ++j)
      n[i][j] = m[i][j] - mat.Element(i, j);

  return Matrica3x3(n);
}

__device__ __host__ void Matrica3x3::operator-()
{
  for (int i = 0; i < 3; ++i)
    for (int j = 0; j < 3; ++j)
      m[i][j] = -m[i][j];
}

__device__ __host__ Realan Matrica3x3::Element(int i, int j) const
{
  return m[i][j];
}

__device__ __host__ bool Matrica3x3::operator==(const Matrica3x3 &matB) const
{
  for (int i = 0; i < 3; ++i)
    for (int j = 0; j < 3; ++j)
      if (fabs(matB.Element(i, j) - m[i][j] < epsilon))
	return false;

  return true;
}

__device__ __host__ bool Matrica3x3::operator!=(const Matrica3x3 &matB) const
{
  return !operator==(matB);
}

__device__ __host__ Matrica3x3 operator*(const Matrica3x3 &matA, const Matrica3x3 &matB)//tes const
{
  Realan n[3][3] = {{0}};
  for (int i = 0; i < 3; ++i)
    for (int j = 0; j < 3; ++j)
      for (int k = 0; k < 3; ++k)
#ifdef FP_FAST_FMA
	n[i][j] = fma (matA.Element(i, k), matB.Element(k, j), n[i][j]);
#else
  n[i][j] += matA.Element(i, k) * matB.Element(k, j);
#endif

  return Matrica3x3(n);
}

__device__ __host__ Matrica3x3 operator*(Realan skalar, const Matrica3x3 &mat)
{
  Realan n[3][3];
  for (int i = 0; i < 3; ++i)
    for (int j = 0; j < 3; ++j)
      n[i][j] = mat.Element(i, j) * skalar;

  return Matrica3x3(n);
}

__device__ __host__ Matrica3x3 operator*(const Matrica3x3 &mat, Realan skalar)
{
  return operator*(skalar, mat);//?
}

__device__ __host__ Matrica3x3 Matrica3x3::Transponiraj() const
{
  Realan n[3][3];
  for (int i = 0; i < 3; ++i)
    for (int j = 0; j < 3; ++j)
      n[i][j] = m[j][i];

  return Matrica3x3(n);
}

//inline double Matrica4x4::Minora(int i, int j) const
//{


__device__ __host__ Realan Matrica3x3::Determinanta()
{
  if (determinantaSpremna)
    return determinanta;
  //else

#ifdef FP_FAST_FMA
  m1 = fma (m[1][1], m[2][2], -m[1][2] * m[2][1]);
  m2 = fma (m[1][0], m[2][2], -m[1][2] * m[2][0]);
  m3 = fma (m[1][0], m[2][1], -m[1][1] * m[2][0]);

  determinanta = fma (m[0][0], m1, fma (-m[0][1], m2, m[0][2] * m3));
#else
  m1 = m[1][1] * m[2][2] - m[1][2] * m[2][1];
  m2 = m[1][0] * m[2][2] - m[1][2] * m[2][0];
  m3 = m[1][0] * m[2][1] - m[1][1] * m[2][0];

  determinanta = m[0][0] * m1 - m[0][1] * m2 + m[0][2] * m3;
#endif

  determinantaSpremna = true;
  return determinanta;
}

__device__ __host__ Matrica3x3 Matrica3x3::Invertiraj()
{
  if (!determinantaSpremna)
    Determinanta();

  if (fabs(determinanta) < epsilon)
    {
      return Matrica3x3();
    }
  //else

#ifdef FP_FAST_FMA
  Realan m4 = fma (m[0][1], m[2][2], -m[0][2] * m[2][1]);
  Realan m5 = fma (m[0][0], m[2][2], -m[0][2] * m[2][0]);
  Realan m6 = fma (m[0][0], m[2][1], -m[0][1] * m[2][0]);
  Realan m7 = fma (m[0][1], m[1][2], -m[0][2] * m[1][1]);
  Realan m8 = fma (m[0][0], m[1][2], -m[0][2] * m[1][0]);
  Realan m9 = fma (m[0][0], m[1][1], -m[0][1] * m[1][0]);
#else
  Realan m4 = m[0][1] * m[2][2] - m[0][2] * m[2][1];
  Realan m5 = m[0][0] * m[2][2] - m[0][2] * m[2][0];
  Realan m6 = m[0][0] * m[2][1] - m[0][1] * m[2][0];
  Realan m7 = m[0][1] * m[1][2] - m[0][2] * m[1][1];
  Realan m8 = m[0][0] * m[1][2] - m[0][2] * m[1][0];
  Realan m9 = m[0][0] * m[1][1] - m[0][1] * m[1][0];
#endif

  return (1 / determinanta * Matrica3x3(m1, -m4, m7, -m2, m5, -m8, m3, -m6, m9));
}

__device__ __host__ Matrica3x3 Matrica3x3::Jedinicna()
{
  Realan n[3][3] = {{1.f, 0.f, 0.f}, {0.f, 1.f, 0.f}, {0.f, 0.f, 1.f}};

  return Matrica3x3(n);
}

__device__ __host__ Matrica3x3 Matrica3x3::RotirajX(Realan radijana)
{
  Realan n[3][3] = {{1}};
  //  n[0][0] = 1;
  n[1][1] = cos(radijana);
  n[1][2] = -sin(radijana);
  n[2][1] = sin(radijana);
  n[2][2] = cos(radijana);

  return Matrica3x3(n);
}

__device__ __host__ Matrica3x3 Matrica3x3::RotirajY(Realan radijana)
{
  Realan n[3][3] = {{0}};
  n[0][0] = cos(radijana);
  n[0][2] = sin(radijana);
  n[1][1] = 1;
  n[2][0] = -sin(radijana);
  n[2][2] = cos(radijana);

  return Matrica3x3(n);
}

__device__ __host__ Matrica3x3 Matrica3x3::RotirajZ(Realan radijana)
{
  Realan n[3][3] = {{0}};
  n[0][0] = cos(radijana);
  n[0][1] = -sin(radijana);
  n[1][0] = sin(radijana);
  n[1][1] = cos(radijana);
  n[2][2] = 1;

  return Matrica3x3(n);
}

