#include "OBJ.h"
#include <fstream>
#include <cctype>
//TODO: parsiranje mtllib i materijala iz .mtl-ova

ModelOBJ::ModelOBJ(const std::string &datoteka)
{
  std::ifstream ulaz(datoteka);
  this->ulaz = &ulaz;
  //TODO: neki exception bi bio dobar ako otvaranje nije uspjelo
  Parsiraj();
  this->ulaz = nullptr;
}

template <std::size_t N, class T>
void ModelOBJ::ParsirajOpcionalne(const std::array<T*, N> &args)
{
  for(T *arg : args)
    {
      *ulaz >> std::ws;
      char c = ulaz->peek();
      if(c != EOF && isdigit(c) || c == '-')
	*ulaz >> *arg;
      else
	break;
    }
}

void ModelOBJ::ParsirajIndeksTrojku(int &vi, int &ti, int &ni) //indeksi su u OBJ-u od 1 (-1 relativni), t.d. 0 ovdje znači invalid
{
  ti = 0;
  ni = 0;
  
  *ulaz >> vi;
  *ulaz >> std::ws;
  if(ulaz->peek() == '/')
    {
      ulaz->ignore();
      if(ulaz->peek() != '/' && ulaz->peek() != '\n')
	{
	  *ulaz >> ti;
	}

      *ulaz >> std::ws;
      if(ulaz->peek() == '/')
	{
	  ulaz->ignore();
	  *ulaz >> ni;
	  *ulaz >> std::ws;
	}
    }
}

void ModelOBJ::ParsirajVrh()
{
  ProvjeriObjekt();
  Realan x, y, z, w = 1.0f;
  *ulaz >> x >> y >> z;
  ParsirajOpcionalne<1>({&w});
  objekti.back().vrhovi.push_back(Vektor4{x, y, z, w});
  //for(Grupa *grupa : aktivneGrupe.Daj())
  //grupa->vrhovi.push_back(Vektor4{x, y, z, w});
}

void ModelOBJ::Parsiraj()
{
  std::string tag;

  do
    {
      *ulaz >> tag;
      // if(tag[0] == '#') //preskoči komentar
      // 	{
      // 	  ulaz->seekg(-1, ios_base::cur);
      // 	  ulaz->ignore(4096, '\n');
      // 	}
      
      if(tag == "v")
	{
	  ParsirajVrh();
	}

      else if(tag == "vt")
	  ParsirajUV();

      else if(tag == "vn")
	ParsirajNormalu();

      else if(tag == "f")
	ParsirajPoligon();

      else if(tag == "o")
	ParsirajObjekt();

      else if(tag == "g")
	ParsirajGrupu();

      else if(tag == "s")
	ParsirajS();

      else if(tag == "usemtl")
	ParsirajUsemtl();

      else
	{
	  ulaz->seekg(-1, std::ios_base::cur);
	  ulaz->ignore(4096, '\n');
	}
    }while(ulaz->peek() != EOF && !ulaz->eof());
}

void ModelOBJ::ParsirajS()
{
  std::string aktivno;
  *ulaz >> aktivno;
  if(aktivno != "0" && aktivno != "off")
    aktivneGrupe.IzvediZaSve([&](Grupa *g) {g->smoothGroup = atoi(aktivno.c_str());});
  else
    aktivneGrupe.IzvediZaSve([&](Grupa *g) {g->smoothGroup = 0;});
}

void ModelOBJ::ParsirajUV()
{
  ProvjeriObjekt();
  Realan u, v = 0.0f, w = 0.0f;
  *ulaz >> u;
  ParsirajOpcionalne<2>({&v, &w});
  objekti.back().UV.push_back(Vektor3{u, v, w});
  //aktivneGrupe.IzvediZaSve([&](Grupa *g) {g->UV.push_back(Vektor3{u, v, w});});
}

void ModelOBJ::ParsirajNormalu()
{
  ProvjeriObjekt();
  Realan x, y, z;
  *ulaz >> x >> y >> z;
  objekti.back().normale.push_back(Vektor3{x, y, z});
  //  aktivneGrupe.IzvediZaSve([&](Grupa *g) {g->normale.push_back(Vektor3{x, y, z});});
}

void ModelOBJ::ParsirajPoligon()
{
  ProvjeriGrupu();
  Poligon poligon;
  int *idx = &poligon.a;
  for(int i = 0; i < 3; ++i, idx+=3)
    {
      ParsirajIndeksTrojku(*(idx+0), *(idx+1), *(idx+2));
    }

  aktivneGrupe.IzvediZaSve([&](Grupa *g) {g->poligoni.push_back(poligon);});
}

void ModelOBJ::ProvjeriObjekt()
{
  if(objekti.empty())
    {
      objekti.push_back(Objekt{});
      aktivneGrupe = SkupGrupa{objekti.back().grupe};
    }
}

void ModelOBJ::ProvjeriGrupu()
{
  if(objekti.back().grupe.empty())
    {
      objekti.back().grupe.insert(std::make_pair<std::string, Grupa>
				  ("default", Grupa{}));
      aktivneGrupe.Dodaj("default");
    }
}

void ModelOBJ::ParsirajObjekt()
{
  objekti.push_back(Objekt{});
  aktivneGrupe = SkupGrupa{objekti.back().grupe};
  *ulaz >> objekti.back().naziv;
}

void ModelOBJ::ParsirajGrupu()
{
  char c;
  aktivneGrupe.Reset();
  
  do
    {
      std::string naziv = "";
      do
	{
	  //*ulaz >> naziv;
	  c = ulaz->get();
	  if(!isspace(c))
	    naziv.append(1, c);
	}while(!isspace(c));
      
      aktivneGrupe.Dodaj(naziv);
    }while(c != '\n');
}

void ModelOBJ::ParsirajUsemtl()
{
  ProvjeriGrupu();
  aktivneGrupe.IzvediZaSve([](Grupa *g) {g->materijal = std::shared_ptr<Materijal>(new Materijal);});
  std::string matname;
  *ulaz >> matname;
  //TODO: konstruirati Materijal ovdje preko parsiranja mtl-a!
}


      
