#pragma once
//Čitač OBJ 3d mesh formata
//podržava teksture, trokute, normale
//teksture referenciraju MTL materijal s dodatnim metapodacima
//podržava samo diskretne mesheve; NURBS i ostale krivulje se preskaču
//https://en.wikipedia.org/wiki/Wavefront_.obj_file
#include "MatBaza.h"
#include "Vektor4.h"
#include "Vektor3.h"
#include "Materijal.h"
#include <vector>
#include <unordered_map>
#include <functional>
#include <memory>
#include <fstream>
#include <array>

class ModelOBJ
{
 public:
  ModelOBJ(const std::string &datoteka);

  struct Poligon //pretpostavka: samo trokuti!
  {
    //bool gladak = true; //false bi ovdje značilo da ovo za razliku od cijele grupe ne treba renderirati glatko (iznimka)!
    int a, ta, na;
    int b, tb, nb;
    int c, tc, nc;
  };
  
  struct Grupa
  {
    //std::string naziv;
    int smoothGroup; //0 znači da se ne koristi; umjesto toga koristi vn
    
    size_t brojPoligona;
    std::vector<Poligon> poligoni;
    std::shared_ptr<Materijal> materijal;
  };
  
  struct Objekt
  {
    std::string naziv;
    size_t brojGrupa;
    size_t brojVrhova;
    size_t brojUV;
    size_t brojNormala;
    std::vector<Vektor4> vrhovi;
    std::vector<Vektor3> normale;
    std::vector<Vektor3> UV;
    
    std::unordered_map<std::string, Grupa> grupe;
  };


  const std::vector<Objekt> &DajObjekte() const {return objekti;}

  
 private:
  class SkupGrupa
  {
    std::vector<Grupa*> grupe;
    std::unordered_map<std::string, Grupa> *mapa;

  public:
  SkupGrupa(std::unordered_map<std::string, Grupa> &g) : mapa(&g){}
    SkupGrupa() = default;
    
    void Dodaj(const std::string &naziv)
    {
      if(naziv != "")
	grupe.push_back(&((*mapa)[naziv])); //stvara grupu ako ne postoji!
    }

    void Reset() {grupe.clear();}
    
    std::vector<Grupa*> &Daj() {return grupe;}

    void IzvediZaSve(std::function<void(Grupa*)> fun)
    {
      for(Grupa *grupa : grupe)
	fun(grupa);
    }
    
  } aktivneGrupe;
  
  size_t brojObjekata;
  std::vector<Objekt> objekti;

  std::ifstream *ulaz = nullptr;
  //std::vector<std::string> aktivneGrupe;

  template <std::size_t N, class T = Realan>
    void ParsirajOpcionalne(const std::array<T*, N> &args);

  //auto &DajGrupe() {return objekti.back().grupe.back();} //trenutna (zadnje ubačena) grupa
  void ParsirajIndeksTrojku(int &a, int &b, int &c);
  
  void Parsiraj();
  void ParsirajVrh();
  void ParsirajUV();
  void ParsirajNormalu();
  void ParsirajPoligon();
  void ParsirajObjekt();
  void ParsirajGrupu();
  void ParsirajS();
  void ParsirajUsemtl();

  void ProvjeriObjekt();
  void ProvjeriGrupu();
};

  
    
