#pragma once
#include "Matrica4x4.h"
#include "Svjetlo.h"
#include "Kamera.h"
#include "Zraka.h"
#include "RGB.h"
#include <memory>

/** Osnovna klasa za sve primitivne grafičke objekte koji se mogu direktno renderirati.
 *
 *  Ova klasa bi trebala biti čisto apstraktna, ali budući da ovaj objekt mora doći do GPU-a i tamo izvoditi metode nad raznim objektima, ne možemo imati virtualne metode jer ih HIP trenutno ne podržava.
 */
class ObjektGPU
{
 protected:

 public:
  /** Transformacija iz model prostora u svjetski prostor.
   */
  Matrica4x4 transformacija = Matrica4x4::Jedinicna();

  /** Primjenjuje transformaciju zadanu matricom na objekt.
   */
  __device__ __host__ void Transformiraj(int) {}
  /** Vraća broj niti potreban za 100% paralelnu transformaciju objekta (tj. svih njegovih elemenata) na GPU-u.
   */
  __device__ __host__ int DajBrojNitiZaTransform() const {return 0;}

  __device__ __host__ ObjektGPU() = default;
  __device__ __host__ ~ObjektGPU() = default;

  __device__ __host__ ObjektGPU(const ObjektGPU &) = default;
  __device__ __host__ ObjektGPU & operator=(const ObjektGPU &) = default;
  __device__ __host__ ObjektGPU(ObjektGPU &&) = default;
  __device__ __host__ ObjektGPU & operator=(ObjektGPU &&) = default;

  //  void PostaviTransformaciju(const Matrica4x4 &);

  /** Apstraktna metoda za sjenčanje (shading) objekta u raytracingu.
   *
   *  Podklase će nadjačati (specijalizirati) ovu metodu sa željenim shaderom. U trenutnoj implementaciji, neke primitive mogu odstupiti od ovog prototipa, pa bi trebalo refaktorirati s proširenjem koje dopušta da se bilo koji `int` indeks može vratiti preko reference u ovoj metodi. Na taj način bi se moglo direktno podržati složenije primitive poput indeksiranih mesheva koje zahtijevaju informaciju o točnom trokutu koji je presječen od strane neke zrake.
   *  \param[in] sjeciste  Izračunato sjecište objekta i zrake.
   *  \param[in] svjetla   Polje svjetala na sceni koja utječu na prikaz objekata.
   *  \param[in] n         Broj svjetala u polju.
   *  \return RGB vrijednost koji predstavlja konačnu boju piksela koji odgovara trenutnoj zraci.
   */
  __device__ __host__ RGB Sjencaj(const Vektor3 &sjeciste, const Svjetlo svjetla[], size_t n) {}

  /** Apstraktna metoda za presjecanje zrake i objekta u raytracingu.
   *
   *  Podklase nadjačavaju ovu metodu s točnim (što optimalnijim) algoritmom za presjek zrake i konkretnog geometrijskog objekta (ili jednostavnijeg proxy-a ako je to dovoljno dobro).
   *  \param[in] zraka     Ulazna zraka.
   *  \param[out] t        Parametar \f$t\f$ u jednadžbi zrake koji odgovara pronađenom sjecištu.
   *  \param[out] sjeciste Pronađeno sjecište; podklase ovo mogu koristiti na različite načine.
   *  \return Vraća `true` ako zraka siječe objekte, a `false` inače.
   */
  __device__ __host__ bool PresijeciZraku(const Zraka &zraka,Realan &t,Vektor3 &sjeciste) {}
  //^TODO: ove dvije metode trebaju biti u zasebnoj mixin klasi jer im prototip ovisi o specifičnim zahtjevima: KompozitniModel (treba element id), JednostavanModel
};
