#include "Prozor.h"

Prozor::Prozor(const SDLIntegracija &ctx, std::string naslov, int sirina, int visina) : w(sirina), h(visina), naslov(naslov)
{
  int greska = SDL_CreateWindowAndRenderer(w, h, SDL_WINDOW_RESIZABLE, &prozor, &renderer);
  if(greska == -1)
    std::cout << "Prozor nije mogao biti stvoren!" << std::endl;
  
  else
    {
      SDL_SetWindowTitle(prozor, this->naslov.c_str());
      SDL_SetRenderDrawColor(renderer, 0, 0, 0, 255);
      SDL_RenderClear(renderer);
      SDL_RenderPresent(renderer);

      raster = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_RGB888,
				 SDL_TEXTUREACCESS_STREAMING, w, h);

      
    }
}

Prozor::~Prozor()
{
  Unisti();
}

void Prozor::Unisti()
{
  if(renderer)
    SDL_DestroyRenderer(renderer);

  if(raster)
    SDL_DestroyTexture(raster);

  if(prozor)
    SDL_DestroyWindow(prozor);
  
  renderer = nullptr;
  raster = nullptr;
  prozor = nullptr;
}

void Prozor::UloviEvente()
{
  SDL_Event event;
  while(SDL_PollEvent(&event))//TODO:restrukturirati ovu petlju t.d. radi sa SDL_PeepEvents() i da ne skida evente iz reda; na taj način ostali prozori mogu vidjeti svoje evente bez da ih ovaj makne (alt pristup: supermonitor koji prođe kroz sve evente i pozove odgovarajuće handlere pojedinih prozora kojima pojedini eventi pripadaju)
    {
      switch(event.type)
	{
	case SDL_QUIT:
	  if(izlaz())
	    Unisti();
	  break;

	case SDL_WINDOWEVENT:
	  if(event.window.event == SDL_WINDOWEVENT_RESIZED)
	    {
	      std::pair<int, int> dim = std::make_pair(event.window.data1, event.window.data2);
	      promjenaVelicine(dim);
	      PromjeniVelicinuEvent(dim.first, dim.second);
	    }
	  break;

	case SDL_KEYDOWN:
	  SDL_Keycode tipka = event.key.keysym.sym;
	  stisnutaTipka(tipka);
	  break;
	}
    }
}

void Prozor::Vsync(Uint32 *pikseli)
{
  UloviEvente();
  
  SDL_UpdateTexture(raster, nullptr, pikseli, w * 4);
  SDL_RenderClear(renderer);
  SDL_RenderCopy(renderer, raster, nullptr, nullptr);
  SDL_RenderPresent(renderer);
}

void Prozor::PromjeniVelicinu(int sirina, int visina)
{
  SDL_Event event; //forsiramo resize
  SDL_WindowEvent resizeEvent;
  
  Uint32 wid = SDL_GetWindowID(prozor);
  resizeEvent.windowID = wid;
  resizeEvent.event = SDL_WINDOWEVENT_RESIZED;
  resizeEvent.data1 = sirina;
  resizeEvent.data2 = visina;

  event.type = SDL_WINDOWEVENT;
  event.window = resizeEvent;
  SDL_PushEvent(&event);
}

void Prozor::PromjeniVelicinuEvent(int sirina, int visina)
{
  w = sirina;
  h = visina;

  SDL_DestroyTexture(raster);
  raster = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_RGB888,
			     SDL_TEXTUREACCESS_STREAMING, w, h);
}

void Prozor::PoveziIzlaz(std::function<bool()> izlazHandler)
{
  izlaz = izlazHandler;
}

void Prozor::PoveziPromjenu(std::function<void(std::pair<int,int>)> promjenaHandler)
{
  promjenaVelicine = promjenaHandler;
}

void Prozor::PoveziTipke(std::function<void(SDL_Keycode)> tipkaHandler)
{
  stisnutaTipka = tipkaHandler;
}
