#pragma once
#include "SDLIntegracija.h"
#include "MatBaza.h"
#include "Kanal.h"
#include <cstdint>

/** Tip koji predstavlja općenitu RGB boju.
 *
 *  Interno, boje se predstavljaju kao realan broj, ali na kraju se uvijek koriste u cjelobrojnom obliku unutar fiksnog raspona. Za standardan dinamički raspon (trenutno jedini podržan), taj raspon je \f$[0,255]\f$ i on se koristi pri računanju konačne boje pojedinačnih piksela slike za framebuffer ili nešto sl.
 */
class RGB//8-bitne komponente
{
  //Kanal8 crvena, zelena, plava;
  Kanal8 rgb[3];

 public:
  __device__ __host__ RGB(Kanal8 r, Kanal8 g, Kanal8 b) : rgb{r, g, b} {}
  /*
  __device__ __host__ RGB(Realan r, Realan g, Realan b)
  {
    crvena = static_cast<uint8_t>(r * 255);
    zelena = static_cast<uint8_t>(g * 255);
    plava = static_cast<uint8_t>(b * 255);
  }
  */
  
  __device__ __host__ RGB() = default;
  __device__ __host__ ~RGB() = default;

  __device__ __host__ RGB(const RGB &) = default;
  __device__ __host__ RGB & operator=(const RGB &) = default;
  __device__ __host__ RGB(RGB &&) = default;
  __device__ __host__ RGB & operator=(RGB &&) = default;

  /** Množi RGB boju sa drugom danom bojom.
   *
   *  Ovo se svodi na jednostavno komponentno množenje i ima određenu (približno) fizičku interpretaciju u različitim modelima sjenčanja (npr. Cook-Torrance, Phong itd.).
   *  \param[in] rgb2 Druga boja.
   *  \return Produktna boja.
   */
  __device__ __host__ RGB operator*(const RGB &rgb2)
    {
      RGB nrgb(*this);
      nrgb[0] *= rgb2[0];
      nrgb[1] *= rgb2[1];
      nrgb[2] *= rgb2[2];

      return nrgb;
    }

  /** Daje \f$i\f$-tu komponentu boje.
   *
   *  \f$i\in\{0,1,2\}\f$
   *  \param[in] i Indeks.
   *  \return Odgovarajuća komponenta (kanal) boje.
   */
  __device__ __host__ Kanal8 operator[](int i) const
  {
    return rgb[i];
    //const Kanal8 *pKanal = &crvena;
    //return *(pKanal + idx);
  }

  __device__ __host__ Kanal8& operator[](int idx)
  {
    return rgb[idx];
  }

  /** Pretvara RGB boju u jedan 32-bitni broj.
   *
   *  Slijedom SDL-ovog standardnog formata RGB piksela, crveni kanal ide na najniži bajt.
   */
  __device__ __host__ operator Uint32()
  {
    return (((((0 << 8) | static_cast<uint8_t>(rgb[0])) << 8)
	     | static_cast<uint8_t>(rgb[1])) << 8)
      | static_cast<uint8_t>(rgb[2]);
  }

  __device__ __host__ Kanal8 dajR() const {return rgb[0];}
  __device__ __host__ Kanal8 dajG() const {return rgb[1];}
  __device__ __host__ Kanal8 dajB() const {return rgb[2];}

  __device__ __host__ void postaviR(Kanal8 r) {rgb[0] = r;}
  __device__ __host__ void postaviG(Kanal8 g) {rgb[1] = g;}
  __device__ __host__ void postaviB(Kanal8 b) {rgb[2] = b;}
};
