#include "Scena.h"

std::forward_list<Svjetlo> &Scena::DajSvjetla()
{
  return svjetla;
}

std::forward_list<SferaGPU> &Scena::DajSfere()
{
  return sfere;
}

std::forward_list<TrokutModel> &Scena::DajModele()
{
  return modeli;
}

std::forward_list<TrokutGPU> &Scena::DajTrokute()
{
  return trokuti;
}

int Scena::DajBrojSvjetala() const
{
  return brojSvjetala;
}

int Scena::DajBrojSfera() const
{
  return brojSfera;
}

int Scena::DajBrojModela() const
{
  return brojModela;
}

int Scena::DajBrojTrokuta() const
{
  return brojTrokuta;
}


void Scena::TransformirajObjekte()
{
  for (auto &svjetlo : svjetla)
    {
      svjetlo.Transformiraj();
    }
  
  for (auto &objekt : sfere)
  {
    objekt.Transformiraj(0);
  }

  for (auto &objekt : modeli)
    {
      objekt.Transformiraj(0);
    }

  for (auto &objekt : trokuti)
    {
      objekt.Transformiraj(0);
    }
}


void Scena::DodajSvjetlo(const Svjetlo &svjetlo)
{
  svjetla.push_front(svjetlo);
  ++brojSvjetala;
}

void Scena::DodajSferu(const SferaGPU &sfera)
{
  sfere.push_front(sfera);
  ++brojSfera;
}

int Scena::DodajModel(TrokutModel &&model)
{
  modeli.push_front(std::move(model));
  std::cout << "Dodaj model:"<<modeli.front().podaci.transformacija.Element(0,0)<<'\n';
  return ++brojModela; //TODO: neka svi tipovi prate ovu shemu - vraćamo ID koji se jedini može koristiti za daljnje izmjene/brisanje objekta - to tada uključuje promjenu i host i gpu kopije modela na odgovarajući način
}

void Scena::DodajTrokut(const TrokutGPU &trokut)
{
  trokuti.push_front(trokut);
  ++brojTrokuta;
}

Scena::~Scena()
{
  //sve primitive osim TrokutModel (meshevi) ne zahtijevaju nikakvu posebnu
  //dealokaciju, pa ih ovdje preskačemo
  for(auto &model : modeli)
    {
      std::cout << "Uništavanje CPU modela...\n";
      model.Unisti();
    }
}
