#pragma once
#include "ObjektGPU.h"
#include "SferaGPU.h"
#include "TrokutModel.h"
#include "TrokutGPU.h"
#include <forward_list>
#include <memory>

/** Osnovna klasa za scenu tj. kolekciju svih objekata za renderirati.
 *
 *  Korisnik bi uvijek trebao raditi samo s objektima ovog tipa kada želi nešto renderirati i smjestiti objekte. Instance drže svoje podatke isključivo u host memoriji; za rad s GPU-om je namijenjena posebna klasa ScenaGPU koja može instancirati svoje objekte koristeći upravo instance ove klase.
 *
 *  Za CPU raytracing, potrebno je direktno koristiti objekte ove klase i prilagođene rutine za raytracing; to bi trebalo unificirati u jednu zajedničku strukturu, no budući da na device kodu ne možemo koristiti proizvoljne STL strukture, taj pristup je dosta ograničavajuć i spor za razvoj pa je zasad ovo riješeno duplikacijom klasa ovisno jesu li namijenjene specifično CPU ili specifično GPU renderiranju.
 */
class Scena
{
 private:
  int brojSvjetala = 0;
  int brojSfera = 0;
  int brojModela = 0;
  int brojTrokuta = 0;
//  std::forward_list<std::unique_ptr<Svjetlo>> svjetla;
//  std::forward_list<std::unique_ptr<Objekt>> objekti;
  std::forward_list<Svjetlo> svjetla;
  std::forward_list<SferaGPU> sfere;
  std::forward_list<TrokutModel> modeli;
  std::forward_list<TrokutGPU> trokuti;

 public:
  Scena() = default;
  ~Scena();

  void TransformirajObjekte();

  std::forward_list<Svjetlo> &DajSvjetla();
  std::forward_list<SferaGPU> &DajSfere();
  std::forward_list<TrokutModel> &DajModele();
  std::forward_list<TrokutGPU> &DajTrokute();

  int DajBrojSvjetala() const;
  int DajBrojSfera() const;
  int DajBrojModela() const;
  int DajBrojTrokuta() const;

  void DodajSvjetlo(const Svjetlo&);
  void DodajSferu(const SferaGPU&);
  int DodajModel(TrokutModel&&); //TODO: unificiraj sve tipove objekata na ovaj način i preko GPUCopyable interfacea i vezanog načina imenovanja parova CPU-GPU klasa
  void DodajTrokut(const TrokutGPU&);
};
