#include "ScenaGPU.h"

__host__ ScenaGPU::ScenaGPU(Scena &scena)
{
  size_t svjetlaBajtova = sizeof(Svjetlo) * scena.DajBrojSvjetala();
  size_t sferaBajtova = sizeof(SferaGPU) * scena.DajBrojSfera();
  size_t modeliBajtova = sizeof(TrokutModelBaza::Podaci) * scena.DajBrojModela();
  size_t trokutiBajtova = sizeof(TrokutGPU) * scena.DajBrojTrokuta();
  
  if(podaci.svjetla)
    {
      hipFree(podaci.svjetla);
    }
  
  hipMalloc(&podaci.svjetla, svjetlaBajtova);

  if(podaci.sfere)
    {
      hipFree(podaci.sfere);
    }

  hipMalloc(&podaci.sfere, sferaBajtova);

  if(podaci.modeli)
    {
      hipFree(podaci.modeli);
    }

  hipMalloc(&podaci.modeli, modeliBajtova);

  if(podaci.trokuti)
    {
      hipFree(podaci.trokuti);
    }

  hipMalloc(&podaci.trokuti, trokutiBajtova);

  brojNiti = 0;

  Svjetlo *poljeSvjetla = new Svjetlo[scena.DajBrojSvjetala()];
  int i = 0;
  for(auto &svjetlo : scena.DajSvjetla())
    {
      poljeSvjetla[i++] = svjetlo;
    }

  hipMemcpy(podaci.svjetla, poljeSvjetla, svjetlaBajtova, hipMemcpyHostToDevice);
  delete[] poljeSvjetla;
  
  brojSferaNiti = 0;
  SferaGPU *poljeSfera = new SferaGPU[scena.DajBrojSfera()];
  i = 0;
  for(auto &sfera : scena.DajSfere())
    {
      poljeSfera[i++] = sfera;
      brojNiti += sfera.DajBrojNitiZaTransform();
      brojSferaNiti += sfera.DajBrojNitiZaTransform();
    }

  hipMemcpy(podaci.sfere, poljeSfera, sferaBajtova, hipMemcpyHostToDevice);
  delete[] poljeSfera;

  auto *poljeModela = new TrokutModelBaza::Podaci[scena.DajBrojModela()];
  i = 0;
  brojModelNiti = 0;
  for(auto &model : scena.DajModele())
    {
      poljeModela[i++] = model.PrekopirajNaGPU().DajPodatke();//TODO: neka svi tipovi objekata rade kopiranje preko istog interfacea, na ovaj usuglašen način
      brojNiti += model.DajBrojNitiZaTransform();
      brojModelNiti += model.DajBrojNitiZaTransform();
    }

  hipMemcpy(podaci.modeli, poljeModela, modeliBajtova, hipMemcpyHostToDevice);
  delete[] poljeModela; ///TODO: prebaci sve ovo u jedan memorijski bazen!

  TrokutGPU *poljeTrokuta = new TrokutGPU[scena.DajBrojTrokuta()];
  i = 0;
  brojTrokutNiti = 0;
  for(auto &trokut : scena.DajTrokute())
    {
      poljeTrokuta[i++] = trokut;
      brojNiti += trokut.DajBrojNitiZaTransform();
      brojTrokutNiti += trokut.DajBrojNitiZaTransform();
    }

  hipMemcpy(podaci.trokuti, poljeTrokuta, trokutiBajtova, hipMemcpyHostToDevice);
  delete[] poljeTrokuta;

  //TODO: refaktorirati sferu i trokut da rade s GPUResident interfaceom koji nam daje pogodnu apstrkaciju za pristup svim podacima koji su u svim tipovima uvijek TriviallyCopyable

  podaci.brojSvjetala = scena.DajBrojSvjetala();
  podaci.brojSfera = scena.DajBrojSfera();
  podaci.brojTrokutModela = scena.DajBrojModela();
  podaci.brojTrokuta = scena.DajBrojTrokuta();
  std::cout << "brojTrokutModela: " << podaci.brojTrokutModela << std::endl;
}

__device__ __host__ void ScenaGPU::TransformirajObjekte(ScenaPodaci &podaci)
{
  for(int i = 0; i < podaci.brojSvjetala; ++i)
    podaci.svjetla[i].Transformiraj();

  for(int i = 0; i < podaci.brojSfera; ++i)
    podaci.sfere[i].Transformiraj(0);


  for(int i = 0; i < podaci.brojTrokuta; ++i)
    podaci.trokuti[i].Transformiraj(0);
 
}

__host__ ScenaGPU &ScenaGPU::operator=(const ScenaGPU &&scena2)
{
  /*
  if(podaci.svjetla)
    hipFree(podaci.svjetla);

  if(podaci.sfere)
    hipFree(podaci.sfere);

  if(podaci.modeli)
    hipFree(podaci.modeli);

  if(podaci.trokuti)
    hipFree(podaci.trokuti);
  */
  Unisti();

  podaci.brojSvjetala = scena2.DajBrojSvjetala();
  podaci.brojSfera = scena2.DajBrojSfera();
  podaci.brojTrokutModela = scena2.DajBrojModela();
  podaci.brojTrokuta = scena2.DajBrojTrokuta();

  podaci.svjetla = scena2.DajSvjetla();
  podaci.sfere = scena2.DajSfere();
  podaci.modeli = scena2.DajModele();
  podaci.trokuti = scena2.DajTrokute();

  return *this;
}

__device__ __host__ size_t ScenaGPU::DajBrojTransformNiti() const
{
  return brojNiti;
}

__host__ void ScenaGPU::Unisti()
{
  if(podaci.svjetla)
    hipFree(podaci.svjetla);
  
  if(podaci.sfere)
    hipFree(podaci.sfere);

  if(podaci.trokuti)
    hipFree(podaci.trokuti);

  //hack za uništavanje objekata s dinamičkom GPU-side alokacijom;
  //ovo treba rješiti sa shared_ptr i prirodno koristeći destruktor,
  //ali HIP to trenutno ne podržava (?)
  if(podaci.modeli)
    {
      std::cout << "ScenaGPU::Unisti entering" << std::endl;
      for(int i = 0; i < podaci.brojTrokutModela; ++i)
	{
	  std::cout << "Pretest\n";
	  TrokutModelBaza::Podaci tmp;
	  hipMemcpy(&tmp, &podaci.modeli[i], sizeof(TrokutModelBaza::Podaci), hipMemcpyDeviceToHost);
	  std::cout << "Post test\n" << &tmp;
	  TrokutModelGPU model(tmp);
	  std::cout << "kopija ok" << std::endl;
	  model.Unisti();  
	}

      hipFree(podaci.modeli);
    } 
}

__host__ ScenaGPU::~ScenaGPU()
{
  Unisti();
}

int ScenaGPU::DajBrojSfera() const
{
  return podaci.brojSfera;
}

int ScenaGPU::DajBrojModela() const
{
  return podaci.brojTrokutModela;
}

int ScenaGPU::DajBrojTrokuta() const
{
  return podaci.brojTrokuta;
}

int ScenaGPU::DajBrojSvjetala() const
{
  return podaci.brojSvjetala;
}

SferaGPU *ScenaGPU::DajSfere() const
{
  return podaci.sfere;
}

TrokutModelBaza::Podaci *ScenaGPU::DajModele() const
{
  return podaci.modeli;
}

TrokutGPU *ScenaGPU::DajTrokute() const
{
  return podaci.trokuti;
}

Svjetlo *ScenaGPU::DajSvjetla() const
{
  return podaci.svjetla;
}

/*
void ScenaGPU::TransformirajObjekte()
{
  for (auto &objekt : sfere)
  {
    objekt->Transformiraj();
  }

  for (auto &objekt : kvadrovi)
    {
      objekt->Transformiraj();
    }

  for (auto &objekt : trokuti)
    {
      objekt->Transformiraj();
    }
}
*/
