#include "SferaGPU.h"
#include "Kanal.h"

__device__ SferaGPU::SferaGPU(Realan r) : R(r)
{
  //hipMalloc(&dR, sizeof(Realan));
  //hipMalloc(&drR, sizeof(Realan));
  
  reciprocR = 1.0 / R;
  //hipMemcpy(dR, &r, sizeof(Realan), hipMemcpyHostToDevice);
  //hipMemcpy(drR, &rR, sizeof(Realan), hipMemcpyHostToDevice);
  //dR = r;
  //drR = rR;
}

__device__ SferaGPU::SferaGPU(Realan r, RGB rgb) : SferaGPU(r)
{
  kd = rgb;
}

__device__ __host__ void SferaGPU::Transformiraj(int offset)
{
}

__host__ __device__ bool SferaGPU::PresijeciZraku(const Zraka &zraka, Realan &t, Vektor3 &sjeciste)
{
  Vektor3 c = Vektor3(transformacija.Element(0, 3), transformacija.Element(1, 3), transformacija.Element(2, 3));
  Vektor3 e = zraka.Ishodiste();
  Vektor3 d = zraka.Smjer();
  Vektor3 u = e - c;
  Realan D = pow(d * u, 2) - (d * d) * (u * u - R * R);

  if (D < 0)
    return false;

  //  Realan t1 = -(d * u + MatBaza::sign(d * u) * sqrt(D)) / (d * d);
  Realan t1 = -(d * u - copysign(sqrt(D), d * u)) / (d * d);

  sjeciste = e + d * t1;
  t = t1;
  
  return true;
}

__host__ __device__ RGB SferaGPU::Sjencaj(const Vektor3 &sjeciste, const Svjetlo svjetla[], size_t n)
{
  Vektor3 c = Vektor3(transformacija.Element(0, 3), transformacija.Element(1, 3), transformacija.Element(2, 3));
  Vektor3 normala = (sjeciste - c) * reciprocR;
  //  normala = normala.Normiraj();
  Kanal8 r = 0u, g = 0u, b = 0u;
  for(int i = 0; i < n; ++i)
  {
    const Svjetlo &svjetlo = svjetla[i];
    Vektor3 l = (svjetlo.DajPoziciju().Radijusvektor() - sjeciste).Normiraj();
    Realan m = fmaxf(0, normala * l);
    r += kd[0] * (svjetlo.boja[0] * Kanal8(m));
    g += kd[1] * (svjetlo.boja[1] * Kanal8(m));
    b += kd[2] * (svjetlo.boja[2] * Kanal8(m));
  }

  return RGB{r, g, b};
}

__host__ RGB SferaGPU::Sjencaj(const Vektor3 &sjeciste, const std::forward_list<Svjetlo> &svjetla)
{
  Vektor3 c = Vektor3(transformacija.Element(0, 3), transformacija.Element(1, 3), transformacija.Element(2, 3));
  Vektor3 normala = (sjeciste - c) * reciprocR;
  //normala = normala.Normiraj();
  Kanal8 r = 0u, g = 0u, b = 0u;
  for(const Svjetlo &svjetlo : svjetla)
    {
      Vektor3 l = (svjetlo.DajPoziciju().Radijusvektor() - sjeciste).Normiraj();
      Realan m = fmax(0, normala * l);
      r += kd[0] * (svjetlo.boja[0] * Kanal8(m));
      g += kd[1] * (svjetlo.boja[1] * Kanal8(m));
      b += kd[2] * (svjetlo.boja[2] * Kanal8(m));
    }

  return RGB{r, g, b};
}
