#pragma once
#include "ObjektGPU.h"

/** Tip grafičke primitive za raytracing koji predstavlja sferu.
 *
 *  Ova primitiva je vrlo jednostavna i ne zahtijeva nikakvu dodatnu dinamičku alokaciju memorije, pa se ista klasa koristi i za GPU i CPU raytracing.
 */
class SferaGPU : public ObjektGPU
{
 private:
  /** Radijus sfere.
   */
  Realan R;
  /** Recipročan radijus.
   *
   *  Koristi se za račun sjecišta; detalji se mogu pronaći u seminarskom radu u sekciji *Presjecanje zraka i objekata*.
   */
  Realan reciprocR;
  /** Difuzni koeficijenti.
   *
   *  Ovo se u Phongovom modelu sjenčanja koristeći najjednostavniju Lambertovu formulu može interpretirati kao površinska boja.
   */
  RGB kd = {1.0f, 1.0f, 1.0f};

 public:
  __device__ __host__ SferaGPU() = default;
  /** Konstruktor sfere s radijusom i bojom.
   *  \param[in] r    Radijus sfere.
   *  \param[in] rgb  Površinska boja sfere.
   */
  __device__ __host__ SferaGPU(Realan r, RGB rgb);
  /** Konstruktor sfere s radijusom.
   *  \param[in] r    Radijus sfere.
   */
  __device__ __host__ SferaGPU(Realan r);
  __device__ __host__ ~SferaGPU() = default;

  __device__ __host__ SferaGPU(const SferaGPU &) = default;
  __device__ __host__ SferaGPU & operator=(const SferaGPU &) = default;
  __device__ __host__ SferaGPU(SferaGPU &&) = default;
  __device__ __host__ SferaGPU & operator=(SferaGPU &&) = default;

  __device__ __host__ void Transformiraj(int) override;
  /** Vraća broj niti za transformaciju sfere: 1.
   *
   *  Sfere se ne moraju posebno transformirati, zato je ovo nepotrebna funkcija. U računu sjecišta zrake i sfere, uzima se u obzir točno središte sfere direktno iz transformacijske matrice.
   */
  __device__ __host__ int DajBrojNitiZaTransform() const override {return 1;}

  __device__ __host__ bool PresijeciZraku(const Zraka &zraka, Realan &t, Vektor3 &sjeciste) override;
  __device__ __host__ RGB Sjencaj(const Vektor3 &sjeciste, const Svjetlo[], size_t) override;

  /** Sjenčanje za CPU raytracing.
   *  \param[in] sjeciste  Izračunato sjecište u svjetskom prostoru.
   *  \param[in] svjetla   Lista svjetala na sceni.
   *  \return RGB boja odgovarajućeg piksela.
   */
  __host__ RGB Sjencaj(const Vektor3 &sjeciste, const std::forward_list<Svjetlo> &svjetla);
};
