#pragma once
#include "Zraka.h"

/** Mixin tip klase za generičnu funkcionalnost sječe trokuta.
 *
 *  Ideja je da pojedine primitive koje imaju potrebu siječi zraku s trokutom naslijede ovu klasu. Ovdje je već implementiran osnovni algoritam koji koristi pojedinačan Trokut, ali moglo bi se koristiti i drugdje. Trenutno je beskorisno jer TrokutModel koristi svoju implementaciju; TODO: refaktoriraj to i napravi da ove metode rade čisto funkcionalno i neka budu statičke (stateless; ništa od međurezultata čuvati u klasi jer to dovodi do data raceova u GPU raytracingu).
 */
class SjeceTrokut
{
 protected:
  Vektor3 aTransformiran, bTransformiran, cTransformiran;
  
 public:
  //očekuje se da podklase iskoriste ovu metodu; ne može biti virtual zbog HIP-ovog ograničenja!
  __host__ __device__ bool PresijeciZrakuTrokut(const Zraka &zraka, Realan &t, Vektor3 &bariSjeciste) const
  {
    const Vektor3 &d = zraka.Smjer();
    const Vektor3 &e = zraka.Ishodiste();
    
    Realan eihf = (aTransformiran[1] - cTransformiran[1]) * d[2] - d[1] * (aTransformiran[2] - cTransformiran[2]);
    Realan gfdi = d[0] * (aTransformiran[2] - cTransformiran[2]) - (aTransformiran[0] - cTransformiran[0]) * d[2];
    Realan dheg = (aTransformiran[0] - cTransformiran[0]) * d[1] - (aTransformiran[1] - cTransformiran[1]) * d[0];
    Realan M = (aTransformiran[0] - bTransformiran[0])*eihf + (aTransformiran[1] - bTransformiran[1])*gfdi + (aTransformiran[2] - bTransformiran[2])*dheg;
    
    Realan beta = ((aTransformiran[0] - e[0])*eihf + (aTransformiran[1] - e[1])*gfdi + (aTransformiran[2] - e[2])*dheg) / M;
    Realan gama = ( d[2]*((aTransformiran[0]-bTransformiran[0])*(aTransformiran[1]-e[1]) - (aTransformiran[0]-e[0])*(aTransformiran[1]-bTransformiran[1])) + d[1]*((aTransformiran[0]-e[0])*(aTransformiran[2]-bTransformiran[2]) - (aTransformiran[0]-bTransformiran[0])*(aTransformiran[2]-e[2])) + d[0]*((aTransformiran[1]-bTransformiran[1])*(aTransformiran[2]-e[2])-(aTransformiran[1]-e[1])*(aTransformiran[2]-bTransformiran[2])) ) / M;
  if(beta < 0 || gama < 0 || beta + gama >= 1.0f)
    return false;
  
  t = -( (aTransformiran[2]-cTransformiran[2])*((aTransformiran[0]-bTransformiran[0])*(aTransformiran[1]-e[1]) - (aTransformiran[0]-e[0])*(aTransformiran[1]-bTransformiran[1])) + (aTransformiran[1]-cTransformiran[1])*((aTransformiran[0]-e[0])*(aTransformiran[2]-bTransformiran[2]) - (aTransformiran[0]-bTransformiran[0])*(aTransformiran[2]-e[2])) + (aTransformiran[0] - cTransformiran[0])*((aTransformiran[1]-bTransformiran[1])*(aTransformiran[2]-e[2])-(aTransformiran[1]-e[1])*(aTransformiran[2]-bTransformiran[2])) ) / M;
  
  //sjeciste = e + d*t;
  bariSjeciste = Vektor3(0.0f, beta, gama);//vrati baricentrične koordinate sjecišta; alfa je implicitno poznata i tipično se ne koristi direktno
  return true;
  }

  
};
  
