#pragma once
#include "Svjetlo.h"

/** Mixin klasa za osnovno sjenčanje trokuta.
 *
 *  Ova klasa nudi funkcionalnost sjenčanja trokuta bez interpolacije normala, dakle za najjednostavnije pojedinačne trokute koji imaju samo jednu definiranu normalu. Za složenije mesheve će najčešće biti potrebno raditi dodatne interpolacije i imati više ulaznih informacija, pa je i ovo na kraju u principu beskorisno jer ga koristi samo TrokutGPU.
 */
class SjencaTrokut
{
 public:
  __host__ __device__ RGB SjencajTrokut(const Vektor3 &sjeciste, const Svjetlo svjetla[], size_t n) const
  {
    //Lambert:
    Kanal8 r = 0u, g = 0u, b = 0u;
    for(int i = 0; i < n; ++i)
      {
	const Svjetlo &svjetlo = svjetla[i];
	Vektor3 l = (svjetlo.DajPoziciju().Radijusvektor() - sjeciste).Normiraj();
	Realan m = fmaxf(0.0f, transformiranaNormala * l);
	r += kd[0] * (svjetlo.boja[0] * Kanal8(m));
	g += kd[1] * (svjetlo.boja[1] * Kanal8(m));
	b += kd[2] * (svjetlo.boja[2] * Kanal8(m));
      }

    return RGB{r, g, b};
  }

 protected:
  Vektor3 transformiranaNormala;
  RGB kd = {1.0f, 0.5f, 0.5f};
    
};
