#include "Slika.h"
#include "MatBaza.h"
#include "Objekt.h"

Slika::Slika(int sirina, int visina) : sirina(sirina), visina(visina), slika(new Uint32[sirina * visina]){}

//const std::unique_ptr<Uint32[]> &Slika::DajSliku() const {return slika;}
Uint32 *Slika::DajSliku() const {return slika;}

void Slika::Realociraj(const SlikaGPU &slikaGPU)
{
  auto dimenzije = slikaGPU.DajDimenzije();
  int novaSirina = dimenzije.first;
  int novaVisina = dimenzije.second;

  if(novaSirina * novaVisina > sirina * visina)
    {
      if(slika)
	delete[] slika;
      
      slika = new Uint32[novaSirina * novaVisina];
    }

  sirina = novaSirina;
  visina = novaVisina;
  hipMemcpy(slika, slikaGPU.DajSliku(), sizeof(Uint32) * sirina * visina, hipMemcpyDeviceToHost);
  //  std::cout << "Realokacija host framebuffera ok" << std::endl;
}

Slika::Slika(const SlikaGPU &slikaGPU)
{
  Realociraj(slikaGPU);
}

Slika &Slika::operator=(const SlikaGPU &slikaGPU)
{
  /*
  if(slika)
    {
      delete[] slika;
      slika = nullptr;
    }
  */

  Realociraj(slikaGPU);
  return *this;
}

Slika::~Slika()
{
  delete[] slika;
}

__host__ void Slika::Transformiraj(Scena &scena)
{
  for(auto &sfera : scena.DajSfere())
    sfera.Transformiraj(0);

  for(auto &trokut : scena.DajTrokute())
    trokut.Transformiraj(0);

  for(auto &model : scena.DajModele())
    model.Transformiraj(0);
}

__host__ void Slika::Raytrace(Kamera &kamera, Scena &scena)
{
  Transformiraj(scena);
  
  for(int i = 0; i < kamera.sirina; ++i)
    {
      for(int j = 0; j < kamera.visina; ++j)//SDL ekranski koordinatni sustav: (0,0) GORE lijevo
	{
	  Realan u = kamera.l + i + 0.5;
	  Realan v = kamera.b + j + 0.5;

	  Vektor3 smjer, o;
	  switch(kamera.projekcija)//optimizacija: 2 vrste kernela, za 2 vrste kamere jer jedna ne može odjednom postati druga dinamički i ovdje je to suvišno granati!
	    {
	    case Kamera::TipProjekcije::Ortografska:
	      o = kamera.pozicija.Radijusvektor() + u * kamera.u + v * kamera.v;
	      smjer = -kamera.w;
	      break;

	    case Kamera::TipProjekcije::Perspektivna:
	      o = kamera.pozicija.Radijusvektor();
	      smjer = -kamera.d * kamera.w + u * kamera.u + v * kamera.v;
	      break;
	    }
	  
	  Zraka zraka(o, smjer);
	  SferaGPU *prvaSfera = nullptr;
	  TrokutModel *prviKvadar = nullptr;
	  TrokutGPU *prviTrokut = nullptr;
	  Realan najmanji = BESKONACNO;
	  Vektor3 potencijalnoSjeciste, sjeciste;
	  int presjecenTrokut;
	  
	  Realan t = BESKONACNO;
	  for(auto &objekt : scena.DajModele())
	    {
	      if (!(objekt.PresijeciZraku(zraka, t, presjecenTrokut, potencijalnoSjeciste)))
		continue;
	      
	      if (t < najmanji && t > 0)//iza svih i ispred kamere
		{
		  najmanji = t;
		  prviKvadar = &objekt;
		  prvaSfera = nullptr;
		  prviTrokut = nullptr;
		  sjeciste = potencijalnoSjeciste;
		}
	    }
	  
	  for(auto &objekt : scena.DajTrokute())
	    {
	      Realan t;
	      if (!(objekt.PresijeciZraku(zraka, t, potencijalnoSjeciste)))
		continue;
	      
	      if (t < najmanji && t > 0)//iza svih i ispred kamere
		{
		  najmanji = t;
		  prviTrokut = &objekt;
		  prvaSfera = nullptr;
		  prviKvadar = nullptr;
		  sjeciste = potencijalnoSjeciste;
		}
	    }

	  for(auto &objekt : scena.DajSfere())
	    {
	      Realan t;
	      if (!(objekt.PresijeciZraku(zraka, t, potencijalnoSjeciste)))
		continue;
	      
	      if (t < najmanji && t > 0)//iza svih i ispred kamere
		{
		  najmanji = t;
		  prvaSfera = &objekt;
		  prviKvadar = nullptr;
		  prviTrokut = nullptr;
		  sjeciste = potencijalnoSjeciste;
		}
	    }


	  Uint32 piksel;//RGB piksel;
	  if(!prviTrokut && !prviKvadar && !prvaSfera)
	    {
	      smjer = smjer.Normiraj();
	      Realan tmp0 = (smjer[0]+1)/2.0f;
	      Realan tmp1 = (smjer[1]+1)/2.0f;
	      Realan tmp2 = (smjer[2]+1)/2.0f;
	      //piksel = pozadinskaBoja;
	      piksel = RGB{tmp1*0.4f, tmp1*0.5f, tmp1};
	    }
	  
	  else
	    {
	      RGB rgb;
	      if(prviTrokut)
		{
		  //rgb = RGB{1.0f, 1.0f, 1.0f};
		  rgb = prviTrokut->Sjencaj(sjeciste, scena.DajSvjetla());
		}

	      else if(prviKvadar)
		{
		  rgb = prviKvadar->Sjencaj(presjecenTrokut, sjeciste, scena.DajSvjetla());
	    
		}

	      else if(prvaSfera)
		{
		  rgb = prvaSfera->Sjencaj(sjeciste, scena.DajSvjetla());
		}
	      
	      piksel = rgb;
	    }
	  
	  slika[i + kamera.sirina * (kamera.visina - 1 - j)] = piksel;
	}
    }
}
