#pragma once
#include "Pravac.h"
#include "Kamera.h"
#include "Scena.h"
#include "SlikaGPU.h"
#include "RGB.h"
#include "SDLIntegracija.h"
#include <memory>

/** Glavna klasa za host framebuffer i renderiranje.
 *
 *  Ova klasa je namijenjena za konačno prenošenje renderirane slike s GPU-a na host i konačno prikazvianje unutar prozora. Također, ako se uopće ne koristi GPU već samo CPU za renderiranje, može biti korištena samo ova klasa, bez SlikaGPU.
 */
class Slika
{
 private:
  //const Kamera &kamera;//ref
  //const Scena &scena;//ref
  //std::unique_ptr<Uint32[]> slika;
  /** Interno polje piksela za sliku.
   */
  Uint32 *slika = nullptr;
  /** Dimenzije slike.
   */
  int sirina = 0, visina = 0;

  /** Interna pomoćna funkcija za promjenu slike.
   *
   *  Kada se mijenja rezolucija, potrebno je realocirati memoriju i ova funkcija pomaže u svim tim situacijama i još možda nekim dodatnim.
   */
  void Realociraj(const SlikaGPU &);
  /** Transformira sve objekte na sceni na CPU-u.
   *
   *  Sporo, sekvencijalno transformiranje svih objekata. Napomenimo da to može značiti stotine tisuća iteracija množenja matrica i vektora u slučaju malo složenijih trokutastih mesheva! Tipično se sve ovo radi na GPU-u, pa SlikaGPU radi upravo tako u slučaju GPU raytracinga. Ovo služi samo kao fallback za CPU renderiranje.
   *  \param[in] scena Ulazna scena za transformirati.
   */
  __host__ void Transformiraj(Scena &scena);
  
 public:
  Uint32 pozadinskaBoja=0;
  //RGB pozadinskaBoja = {0, 0, 0};
  //Slika(const Kamera &, const Scena &);
  /** Konstruktor slike s dimenzijama.
   */
 Slika(int sirina, int visina);
 ~Slika();
 Slika(const SlikaGPU &);
 Slika(Slika &) = delete;
 Slika& operator=(const SlikaGPU &);
 
 Slika & operator=(Slika &) = delete;
 Slika(Slika &&) = delete;
 Slika & operator=(Slika &&) = delete;
 
 /** Vraća pokazivač na framebuffer.
  */
 Uint32 *DajSliku() const;

 /** Raytrace na CPU-u.
  *
  *  Renderira scenu s mjesta kamere tehnikom raytracinga na CPU-u.
  *  \param[in] kamera  Kamera koju treba koristiti u renderiranju.
  *  \param[in] scena   Scena za renderirati.
  */
 __host__ void Raytrace(Kamera &kamera, Scena &scena);//renderiraj scenu s mjesta kamere tehnikom raytracinga na CPU-u
};
