#pragma once
//#include "Pravac.h"
#include "Kamera.h"
#include "ScenaGPU.h"
#include "RGB.h"
#include "SDLIntegracija.h"
#include <cstdint>

/** Klasa koja predstavlja sliku na GPU-u.
 *
 *  U ovoj klasi se sve dinamički alocira na GPU-u, a kasnije po potrebi slika prekopira na host radi prikazivanja ili neke daljnje obrade. Ovdje se također događa GPU raytracing i transformacija preko zasebnih kernela.
 */
class SlikaGPU
{
 private:
  //Kamera kamera;
  //const ScenaGPU scena;
  Uint32 *slika = nullptr;
  uint16_t sirina, visina;
  static constexpr Uint32 pozadinskaBoja = 0;
  static constexpr int BLOK_X = 64;
  static constexpr int BLOK_Y = 4; //možda bi trebalo imati manje niti sveukupno, eksperimentiraj... 64 po x je preferabilno radi memory coalescing pri zapisivanju piksela u redu unutar jednog warpa
  static constexpr int MAX_NITI_PO_BLOKU = 1024;
  static constexpr int MAX_BLOKOVA = 8096; //za transformacijski kernel//TODO: stavi ovo kao -D da možemo kompajlirati na prezentaciji za točno optimalan broj?
  /** Broj blokova niti po X odnosno Y osi za renderiranje tražene rezolucije.
   */
  uint16_t brojBlokX, brojBlokY;

  /** GPU transformacijski kernel za scenu.
   *
   *  Ovaj kernel paralelno transformira sve objekte na sceni.
   *  \param[in] scena     Podaci dovoljni za opis GPU scene.
   *  \param[in] offset    Pomak koji treba dodati na trenutni indeks niti kako bi se dobio točan indeks polja pojedinih elemenata objekata na sceni. Potrebno ako je potrebno raditi nekoliko iteracija kernela zbog prevelikog broja elemenata (to se ne bi trebalo uopće događati u praksi).
   *  \param[in] brojSferaNiti   Broj transformacijskih niti za sfere na sceni.
   *  \param[in] brojTrokutNiti  Broj transformacijskih niti za trokute na sceni.
   *  \param[in] brojModelNiti   Broj transformacijskih niti za trokutaste mesheve (indeksirane modele) na sceni.
   */
  static __global__ void transformKernel(ScenaPodaci scena,size_t offset,size_t brojSferaNiti,size_t brojTrokutNiti,size_t brojModelNiti);
  /** GPU raytracing kernel.
   *
   *  \param[in] scena   GPU transformirana scena za renderirati.
   *  \param[in] kamera  Kamera s čijeg stajališta renderirati.
   *  \param[out] slika  Pokazivač na sliku koji se koristi za iscrtavanje.
   */
  static __global__ void rayKernel(ScenaPodaci scena, Kamera kamera, Uint32 *slika);
  /** Ovo se više ne koristi.
   */
  static __host__ void rayCPU(ScenaPodaci scena, Kamera kamera, Uint32 *slika);

  /** Ovo se više ne koristi.
   *
   *  Sva svjetla se transformiraju zajedno s drugim objektima u transformKernelu.
   */
  __host__ void TransformirajSvjetla(ScenaPodaci &scena);

 public:
  //__host__ SlikaGPU(const Kamera &, const ScenaGPU &);
 __host__ SlikaGPU(int sirina, int visina);
  __host__ ~SlikaGPU();
  __host__ SlikaGPU(const SlikaGPU &) = default;
  SlikaGPU & operator=(const SlikaGPU &);
  SlikaGPU & operator=(SlikaGPU &&);
  ///SlikaGPU(SlikaGPU &&); //TODO: implementiraj ovo!

  /** Daje indeksirani piksel slike.
   *  \param[in] idx Indeks
   *  \return RGB piksel pakiran u 32-bitni broj.
   */
  __device__ __host__ Uint32 &operator[](int idx)
  {
    return slika[idx];
  }

  __device__ __host__ void PostaviSliku(Uint32 *nova) {slika = nova;}
  __device__ __host__ Uint32 *DajSliku() const {return slika;}
  __host__ void DajSliku(Uint32 *dest) const;
  /** \return Vraća par (širina,visina) slike.
   */
  __device__ __host__ std::pair<int, int> DajDimenzije() const;
  /** \return Vraća par (bx,by) za broj blokova po pojedinoj dimenziji slike.
   */
  __device__ __host__ std::pair<int, int> DajBrojBlokova() const;

  /** Glavna host funkcija za početak GPU raytracinga.
   *
   *  Ova funkcija se brine o transformaciji, idealnoj raspodjeli niti u blokove i pozive svih odgovarajućih kernela koliko god puta je to potrebno.
   *  \param[in] scena   Scena za renderirati.
   *  \param[in] kamera  Kamera s čijeg stajališta treba renderirati scenu.
   */
  __host__ void Raytrace(ScenaGPU &scena, Kamera &kamera);//renderiraj scenu s mjesta kamere tehnikom raytracinga

  /** Ovo se ne koristi.
   */
  __host__ void RaytraceCPU(ScenaGPU &scena, Kamera &kamera, Uint32 *slika);
};
