#include "Trokut.h"
#include "Kanal.h"
#include <cmath>

//Trokut::Trokut()
//{}

Trokut::Trokut(const Vektor4 &a, const Vektor4 &b, const Vektor4 &c, RGB boja)
{
  this->a = a;
  this->b = b;
  this->c = c;
  normala = ((this->b - this->a).Radijusvektor() ^ (this->c - this->a).Radijusvektor()).Normiraj();

  kd = Vektor3(boja[0] / Kanal8(255.0f), boja[1] / Kanal8(255.0f), boja[2] / Kanal8(255.0f));
  //kd = boja;
}

Trokut::Trokut(const Vektor4 &a, const Vektor4 &b, const Vektor4 &c)
{
  this->a = a;
  this->b = b;
  this->c = c;
  normala = ((this->b - this->a).Radijusvektor() ^ (this->c - this->a).Radijusvektor()).Normiraj();
}

void Trokut::Transformiraj()
{
  aTransformiran = (transformacija * a).Radijusvektor();
  bTransformiran = (transformacija * b).Radijusvektor();
  cTransformiran = (transformacija * c).Radijusvektor();

  transformiranaNormala = (transformacija.Linearna() * normala).Normiraj();//pretpostavka: transformacijska matrica je ortogonalna
}

//bool Trokut::PresijeciZraku(const Zraka &zraka, RGB &boja) const
bool Trokut::PresijeciZraku(const Zraka &zraka, Realan &t)
{
  Vektor3 d = zraka.Smjer();
  Vektor3 e = zraka.Ishodiste();
  //Vektor3 d;
  //Vektor3 e;

//  Vektor3 aTransformiran = (transformacija * a).Radijusvektor();
//  Vektor3 bTransformiran = (transformacija * b).Radijusvektor();
//  Vektor3 cTransformiran = (transformacija * c).Radijusvektor();
  Vektor3 ba = aTransformiran - bTransformiran;
  Vektor3 ca = aTransformiran - cTransformiran;
  Realan a11 = ba[0];
  Realan a21 = ba[1];
  Realan a31 = ba[2];
  Realan a12 = ca[0];
  Realan a22 = ca[1];
  Realan a32 = ca[2];
  /*
  double a11 = a[0] - b[0];
  double a21 = a[1] - b[1];
  double a31 = a[2] - b[2];
  double a12 = a[0] - c[0];
  double a22 = a[1] - c[1];
  double a32 = a[2] - c[2];
  */

  Realan a13 = d[0];
  Realan a23 = d[1];
  Realan a33 = d[2];

  Vektor3 ea = aTransformiran - e;
  Realan b1 = ea[0];
  Realan b2 = ea[1];
  Realan b3 = ea[2];
  /*greška: e!
  double b1 = a[0] - zraka.smjer[0];
  double b2 = a[1] - zraka.smjer[1];
  double b3 = a[2] - zraka.smjer[2];
  */

#ifdef FP_FAST_FMA
  Realan t1 = fma (a22, a33, -a23 * a32);
  Realan t2 = fma (a13, a32, -a12 * a33);
  Realan t3 = fma (a12, a23, -a22 * a13);

  Realan M = fma (a11, t1, fma (a21, t2, a31 * t3));
#else
  Realan t1 = a22 * a33 - a23 * a32;//ei-hf
  Realan t2 = a13 * a32 - a12 * a33;//gf-di
  Realan t3 = a12 * a23 - a22 * a13;//dh-eg

  Realan M = a11 * t1 + a21 * t2 + a31 * t3;
#endif

  if (M == 0.0)
    return false; //ako je normala trokuta ortogonalna zraci, ne vidimo trokut
  //else
  Realan reciprocM = 1 / M;

#ifdef FP_FAST_FMA
  Realan t4 = fma (a11, b2, -b1 * a21);
  Realan t5 = fma (b1, a31, -a11 * b3);
  Realan t6 = fma (a21, b3, -b2 * a31);

  //double beta = (fma (b1, t1, fma (b2, t2, b3 * t3))) * reciprocM;
  //double gama = (fma (a33, t4, fma (a23, t5, a13 * t6))) * reciprocM;

  t = -(fma (a32, t4, fma (a22, t5, a12 * t6))) * reciprocM;
#else

  Realan t4 = a11 * b2 - b1 * a21;//ak-jb
  Realan t5 = b1 * a31 - a11 * b3;//jc-al
  Realan t6 = a21 * b3 - b2 * a31;//bl-kc

  t = -(a32 * t4 + a22 * t5 + a12 * t6) * reciprocM;
#endif

  if(t < 0)
    return false;

#ifdef FP_FAST_FMA
  Realan gama = (fma (a33, t4, fma (a23, t5, a13 * t6))) * reciprocM;
#else
  Realan gama = (a33 * t4 + a23 * t5 + a13 * t6) * reciprocM;
#endif

  if(gama > 1 || gama < 0)
    return false;

#ifdef FP_FAST_FMA
  Realan beta = (fma (b1, t1, fma (b2, t2, b3 * t3))) * reciprocM;
#else
  Realan beta = (b1 * t1 + b2 * t2 + b3 * t3) * reciprocM;
#endif
  if(beta > 1 - gama || beta < 0)
    return false;
//  if (!(beta > 0.0 && gama > 0.0 && ((beta + gama) < 1.0)))
//    return false;
  //else

  //  boja = kd;
  sjeciste = aTransformiran + beta * (bTransformiran - aTransformiran) + gama * (cTransformiran - aTransformiran);
  //  sjeciste = a + beta * (b - a) + gama * (c - a);
  return true;
}

inline RGB Trokut::Sjencaj(const std::forward_list<Svjetlo *> &svjetla)
{
  //Lambert:
  Kanal8 r = 0u, g = 0u, b = 0u;
  for(auto svjetlo : svjetla)
  {
    Vektor3 l = (svjetlo->DajPoziciju().Radijusvektor() - sjeciste).Normiraj();
    Realan m = fmax(0, transformiranaNormala * l);
    r += kd[0] * (svjetlo->boja[0] * Kanal8(m));
    g += kd[1] * (svjetlo->boja[1] * Kanal8(m));
    b += kd[2] * (svjetlo->boja[2] * Kanal8(m));
  }

  return RGB{r, g, b};
//  Kanal8 r, g, b;
//  for(auto svjetlo : svjetla)
//  {
//    Vektor3 l = (svjetlo->DajPoziciju().Radijusvektor() - sjeciste).Normiraj();
//    double m = fmax(0, transformiranaNormala * l);
//    r += kd.crvena * (svjetlo->boja.crvena * m);
//    g += kd.zelena * (svjetlo->boja.zelena * m);
//    b += kd.plava * (svjetlo->boja.plava * m);
//  }

  //return RGB{static_cast<uint8_t>(r), static_cast<uint8_t>(g), static_cast<uint8_t>(b)};
}
