#include "TrokutGPU.h"
#include "Kanal.h"

//Trokut::Trokut()
//{}

__device__ __host__ TrokutGPU::TrokutGPU(const Vektor4 &a, const Vektor4 &b, const Vektor4 &c, RGB boja) :TrokutGPU(a, b, c)
{
  kd = boja;
}

__device__ __host__ TrokutGPU::TrokutGPU(const Vektor4 &a, const Vektor4 &b, const Vektor4 &c) : a(a), b(b), c(c)
{
  normala = (((b - a).Radijusvektor() ^ (c - a).Radijusvektor()).Normiraj());
}

__device__ __host__ TrokutGPU::TrokutGPU(const Vektor4 &a, const Vektor4 &b, const Vektor4 &c, const Vektor3 &n) : a(a), b(b), c(c), normala(n) {}

__device__ __host__ void TrokutGPU::Transformiraj(int offset)
{
  aTransformiran = (transformacija * a).Radijusvektor();
  bTransformiran = (transformacija * b).Radijusvektor();
  cTransformiran = (transformacija * c).Radijusvektor();

  transformiranaNormala = (transformacija.Linearna() * normala).Normiraj();//pretpostavka: transformacijska matrica je ortogonalna; vidite str. 23 rada
}

/*
__device__ __host__ bool TrokutGPU::PresijeciZraku(const Zraka &zraka, Realan &t, Vektor3 &sjeciste) const
{
  const Vektor3 &d = zraka.Smjer();
  const Vektor3 &e = zraka.Ishodiste();
  //ovdje se radi Gaussova eliminacija jer...
  _Float16 A[3][4];
  int8_t i, j, k;

  A[0][0] = aTransformiran[0] - bTransformiran[0];
  A[0][1] = aTransformiran[0] - cTransformiran[0];
  A[0][2] = d[0];
  A[0][3] = aTransformiran[0] - e[0];
  A[1][0] = aTransformiran[1] - bTransformiran[1];
  A[1][1] = aTransformiran[1] - cTransformiran[1];
  A[1][2] = d[1];
  A[1][3] = aTransformiran[1] - e[1];
  A[2][0] = aTransformiran[2] - bTransformiran[2];
  A[2][1] = aTransformiran[2] - cTransformiran[2];
  A[2][2] = d[2];
  A[2][3] = aTransformiran[2] - e[2];
  
  for(i = 0; i < 3; ++i)
    for(j = 0; j < 3; ++j)
      if(fabsf(A[i][i]) < fabsf(A[j][i]))
	{
	  for(k = 0; k < 4; ++k)
	    {
	      A[i][k] += A[j][k];
	      A[j][k] = A[i][k] - A[j][k];
	      A[i][k] = A[i][k] - A[j][k];
	    }
	}

  for(i = 0; i < 2; ++i)
    for(j = i+1; j < 3; ++j)
      {
       _Float16 c = A[j][i] / A[i][i];
	for(k = 0; k < 4; ++k)
	  A[j][k] -= c * A[i][k];
      }

  _Float16 rez[3];
  for(i = 2; i >= 0; i--)
    {
      rez[i] = A[i][3];
      for(j = i+1; j < 3; ++j)
        if(i != j)
	  rez[i] -= A[i][j] * rez[j];
      rez[i] /= A[i][i];
    }

  t = rez[2];
  sjeciste = e + d * t;
  
  if(rez[0] < 0 || rez[1] < 0 || rez[0]+rez[1] >= 1.0f)
    return false;

  return true;
}
*/

__device__ __host__ bool TrokutGPU::PresijeciZraku(const Zraka &zraka, Realan &t, Vektor3 &sjeciste) const
{
  Vektor3 bari;
  if(PresijeciZrakuTrokut(zraka, t, bari))
    {
      sjeciste = aTransformiran + bari[1]*(bTransformiran - aTransformiran) + bari[2]*(cTransformiran - aTransformiran); //vrati viewspace sjecište; vidi str.37 formula (39)
      return true;
    }
  
  else
    return false;

  /*
  const Vektor3 &d = zraka.Smjer();
  const Vektor3 &e = zraka.Ishodiste();

  Realan eihf = (aTransformiran[1] - cTransformiran[1]) * d[2] - d[1] * (aTransformiran[2] - cTransformiran[2]);
  Realan gfdi = d[0] * (aTransformiran[2] - cTransformiran[2]) - (aTransformiran[0] - cTransformiran[0]) * d[2];
  Realan dheg = (aTransformiran[0] - cTransformiran[0]) * d[1] - (aTransformiran[1] - cTransformiran[1]) * d[0];
  Realan M = (aTransformiran[0] - bTransformiran[0])*eihf + (aTransformiran[1] - bTransformiran[1])*gfdi + (aTransformiran[2] - bTransformiran[2])*dheg;

  Realan beta = ((aTransformiran[0] - e[0])*eihf + (aTransformiran[1] - e[1])*gfdi + (aTransformiran[2] - e[2])*dheg) / M;
  Realan gama = ( d[2]*((aTransformiran[0]-bTransformiran[0])*(aTransformiran[1]-e[1]) - (aTransformiran[0]-e[0])*(aTransformiran[1]-bTransformiran[1])) + d[1]*((aTransformiran[0]-e[0])*(aTransformiran[2]-bTransformiran[2]) - (aTransformiran[0]-bTransformiran[0])*(aTransformiran[2]-e[2])) + d[0]*((aTransformiran[1]-bTransformiran[1])*(aTransformiran[2]-e[2])-(aTransformiran[1]-e[1])*(aTransformiran[2]-bTransformiran[2])) ) / M;
  if(beta < 0 || gama < 0 || beta + gama >= 1.0f)
    return false;

  t = -( (aTransformiran[2]-cTransformiran[2])*((aTransformiran[0]-bTransformiran[0])*(aTransformiran[1]-e[1]) - (aTransformiran[0]-e[0])*(aTransformiran[1]-bTransformiran[1])) + (aTransformiran[1]-cTransformiran[1])*((aTransformiran[0]-e[0])*(aTransformiran[2]-bTransformiran[2]) - (aTransformiran[0]-bTransformiran[0])*(aTransformiran[2]-e[2])) + (aTransformiran[0] - cTransformiran[0])*((aTransformiran[1]-bTransformiran[1])*(aTransformiran[2]-e[2])-(aTransformiran[1]-e[1])*(aTransformiran[2]-bTransformiran[2])) ) / M;
  
  sjeciste = e + d*t;
  return true;
  */
}

/*
__device__ __host__ bool TrokutGPU::PresijeciZraku(const Zraka &zraka, Realan &t, Vektor3 &sjeciste) const
{
  Vektor3 stupci[3] = {aTransformiran - bTransformiran, aTransformiran - cTransformiran, d};
  Matrica3x3 mat(stupci);
  _Float16 A = mat.Determinanta();

  stupci[0] = aTransformiran - e;
  stupci[1] = aTransformiran - cTransformiran;
  stupci[2] = d;
  mat = Matrica3x3(stupci);
  _Float16 beta = mat.Determinanta() / A;

  stupci[0] = aTransformiran - bTransformiran;
  stupci[1] = aTransformiran - e;
  stupci[2] = d;
  mat = Matrica3x3(stupci);
  _Float16 gama = mat.Determinanta() / A;

  if(beta < 0 || gama < 0 || beta + gama >= 1)
    return false;

  stupci[0] = aTransformiran - bTransformiran;
  stupci[1] = aTransformiran - cTransformiran;
  stupci[2] = aTransformiran - e;
  mat = Matrica3x3(stupci);
  t = mat.Determinanta() / A;
  sjeciste = e + d * t;

  return true;
#undef aTransformiran
#undef bTransformiran
#undef cTransformiran
}
*/

 /*
__host__ __device__ bool TrokutGPU::PresijeciZraku(const Zraka &zraka, Realan &t, Vektor3 &sjeciste) const
{
  Vektor3 d = zraka.Smjer();
  Vektor3 e = zraka.Ishodiste();
  
  Vektor3 ba = aTransformiran - bTransformiran;
  Vektor3 ca = aTransformiran - cTransformiran;

  _Float16 a11 = ba[0];
  _Float16 a21 = ba[1];
  _Float16 a31 = ba[2];
  _Float16 a12 = ca[0];
  _Float16 a22 = ca[1];
  _Float16 a32 = ca[2];

  _Float16 a13 = d[0];
  _Float16 a23 = d[1];
  _Float16 a33 = d[2];

  Vektor3 ea = aTransformiran - e;
  _Float16 b1 = ea[0];
  _Float16 b2 = ea[1];
  _Float16 b3 = ea[2];

#ifdef FP_FAST_FMA
  _Float16 t1 = fma (a22, a33, -a23 * a32);
  _Float16 t2 = fma (a13, a32, -a12 * a33);
  _Float16 t3 = fma (a12, a23, -a22 * a13);

  _Float16 M = fma (a11, t1, fma (a21, t2, a31 * t3));
#else
  Realan t1 = a22 * a33 - a23 * a32;//ei-hf
  Realan t2 = a13 * a32 - a12 * a33;//gf-di
  Realan t3 = a12 * a23 - a22 * a13;//dh-eg

  Realan M = a11 * t1 + a21 * t2 + a31 * t3;
#endif

  if (M == 0.0)
    return false; //ako je normala trokuta ortogonalna zraci, ne vidimo trokut
  //else
  _Float16 reciprocM = 1 / M;

#ifdef FP_FAST_FMA
  _Float16 t4 = fma (a11, b2, -b1 * a21);
  _Float16 t5 = fma (b1, a31, -a11 * b3);
  _Float16 t6 = fma (a21, b3, -b2 * a31);

  t = -(fma (a32, t4, fma (a22, t5, a12 * t6))) * reciprocM;
#else

  _Float16 t4 = a11 * b2 - b1 * a21;//ak-jb
  _Float16 t5 = b1 * a31 - a11 * b3;//jc-al
  _Float16 t6 = a21 * b3 - b2 * a31;//bl-kc

  t = -(a32 * t4 + a22 * t5 + a12 * t6) * reciprocM;
#endif

  if(t < 0)
    return false;

#ifdef FP_FAST_FMA
  _Float16 gama = (fma (a33, t4, fma (a23, t5, a13 * t6))) * reciprocM;
#else
  _Float16 gama = (a33 * t4 + a23 * t5 + a13 * t6) * reciprocM;
#endif

  if(gama > 1 || gama < 0)
    return false;

#ifdef FP_FAST_FMA
  _Float16 beta = (fma (b1, t1, fma (b2, t2, b3 * t3))) * reciprocM;
#else
  Realan beta = (b1 * t1 + b2 * t2 + b3 * t3) * reciprocM;
#endif
  if(beta > 1 - gama || beta < 0)
    return false;

  sjeciste = aTransformiran + beta * (bTransformiran - aTransformiran) + gama * (cTransformiran - aTransformiran);
  //  sjeciste = atmp + beta * (btmp - atmp) + gama * (ctmp - atmp);
  return true;
}
 */
 
__host__ __device__ RGB TrokutGPU::Sjencaj(const Vektor3 &sjeciste, const Svjetlo svjetla[], size_t n) const
{
  return SjencajTrokut(sjeciste, svjetla, n);
}

__host__ RGB TrokutGPU::Sjencaj(const Vektor3 &sjeciste, const std::forward_list<Svjetlo> &svjetla)
{
  //Lambert:
  Kanal8 r = 0u, g = 0u, b = 0u;
  for(const Svjetlo &svjetlo : svjetla)
      {
	Vektor3 l = (svjetlo.DajPoziciju().Radijusvektor() - sjeciste).Normiraj();
	Realan m = fmaxf(0.0f, transformiranaNormala * l);
	r += kd[0] * (svjetlo.boja[0] * Kanal8(m));
	g += kd[1] * (svjetlo.boja[1] * Kanal8(m));
	b += kd[2] * (svjetlo.boja[2] * Kanal8(m));
      }

  return RGB{r, g, b};
}
