#pragma once
#include "MatBaza.h"
#include "Vektor3.h"
#include "ObjektGPU.h"
#include "SjeceTrokut.h"
#include "SjencaTrokut.h"

/** Tip primitivnog GPU/CPU trokuta.
 *
 *  Ovaj tip je jednostavan i ne zahtijeva korištenje različitih klasa za GPU/CPU renderiranje. Ovaj tip se ne bi trebao koristiti za složene indeksirane mesheve; za to je TrokutModel!
 */
class TrokutGPU : public ObjektGPU, private SjeceTrokut, private SjencaTrokut
{
 private:
  /** Vrhovi trokuta poredani u smjeru obrnutom kazaljci na satu.
   */
  Vektor4 a, b, c;
  /** Vektor normale trokuta.
   */
  Vektor3 normala;
  //Vektor3 aTransformiran, bTransformiran, cTransformiran;
  //Vektor3 transformiranaNormala, normala;

 public:
  __device__ __host__ TrokutGPU() = default;
  /** Konstruktor trokuta s vrhovima i bojom.
   *  
   *  Normala se automatski izračuna.
   *  \param[in] a,b,c Točke vrhova.
   *  \param[in] rgb   RGB boja cijelog trokuta.
   */
  __device__ __host__ TrokutGPU(const Vektor4 &a,const Vektor4 &b,const Vektor4 &c, RGB rgb);
  /** Konstruktor trokuta s vrhovima.
   *
   *  Normala se automatski izračuna.
   *  \param[in] a,b,c Točke vrhova.
   */
  __device__ __host__ TrokutGPU(const Vektor4 &a,const Vektor4 &b,const Vektor4 &c);
  /** Konstruktor trokuta s vrhovima i normalom.
   *  \param[in] a,b,c Točke vrhova.
   *  \param[in] n     Vektor normale.
   */
  __device__ __host__ TrokutGPU(const Vektor4 &a,const Vektor4 &b,const Vektor4 &c, const Vektor3 &n); //pretpostavlja se da je normala jedinični vektor
  
  __device__ __host__ ~TrokutGPU() = default;

  __device__ __host__ TrokutGPU(const TrokutGPU &) = default;
  __device__ __host__ TrokutGPU & operator=(const TrokutGPU &) = default;
  __device__ __host__ TrokutGPU(TrokutGPU &&) = default;
  __device__ __host__ TrokutGPU & operator=(TrokutGPU &&) = default;

  __device__ __host__ void Transformiraj(int) override;
  __device__ __host__ int DajBrojNitiZaTransform() const override {return 1;}

  __host__ __device__ bool PresijeciZraku(const Zraka &,Realan &,Vektor3 &sjeciste) const override;
  __host__ __device__ RGB Sjencaj(const Vektor3 &sjeciste, const Svjetlo[], size_t) const override;

  /** CPU-specifičan shader individualnog trokuta.
   *  \param[in] sjeciste  Izračunato sjecište zrake i trokuta.
   *  \param[in] svjetla   Lista svih svjetala na sceni.
   */
  __host__ RGB Sjencaj(const Vektor3 &sjeciste, const std::forward_list<Svjetlo> &svjetla);
};
