#include "TrokutModel.h"
#include "OBJ.h"

__host__ TrokutModel::TrokutModel(const ModelOBJ::Objekt &obj) //višestruke grupe unutar objekta se ovdje spajaju u jedinstveni mesh; bilo bi korisno izvući bounding box za svaku grupu...
{
  podaci.brojVrhova = obj.vrhovi.size();
  podaci.brojNormala = obj.normale.size();
  podaci.brojUV = obj.UV.size();
  podaci.brojPoligona = 0;

  for(auto &grupa : obj.grupe)
    podaci.brojPoligona += grupa.second.poligoni.size();

  podaci.vrhovi = new Vektor3[podaci.brojVrhova];
  podaci.normale = new Vektor3[podaci.brojNormala];
  podaci.UV = new Vektor3[podaci.brojUV];
  podaci.poligoni = new Poligon[podaci.brojPoligona];
  podaci.transformacija = transformacija;

  podaci.transformiraniVrhovi = new Vektor3[podaci.brojVrhova];
  podaci.transformiraneNormale = new Vektor3[podaci.brojNormala];
  
  Kopiraj(obj);
  //TODO: materijali!
}

__host__ int TrokutModel::SrediIndeks(int idx, int kolicina)
{
  int absIdx;
  if(idx < 0)
    absIdx = kolicina + idx;
  else if(idx == 0)
    return -1;
  else
    absIdx = idx - 1;

  return absIdx;
  //  return startIdx + absIdx;
}

__host__ void TrokutModel::Transformiraj(int nebitno)
{
  for(size_t i = 0; i < podaci.brojVrhova; ++i)
    {
      podaci.transformiraniVrhovi[i] =
	(podaci.transformacija * podaci.vrhovi[i].DajTocku()).Radijusvektor();
    }

  for(size_t i = 0; i < podaci.brojNormala; ++i)
    podaci.transformiraneNormale[i] = podaci.transformacija.Linearna() * podaci.normale[i];
}

__host__ void TrokutModel::Kopiraj(const ModelOBJ::Objekt &obj)
{
  int i = 0;
  for(auto &vrh : obj.vrhovi)
    {
      podaci.vrhovi[i++] = vrh.PerspektivnoDijeljenje();
    }

  static_assert(std::is_trivially_copyable<Vektor3>::value,
		"Vektor3 mora biti TriviallyCopyable za memcpy!");
  std::memcpy(podaci.normale, obj.normale.data(), sizeof(Vektor3) * podaci.brojNormala);
  std::memcpy(podaci.UV, obj.UV.data(), sizeof(Vektor3) * podaci.brojUV);

  i = 0;
  //  int start = 0;
  for(auto &grupa : obj.grupe)
    {
      for(auto &poli : grupa.second.poligoni)
	podaci.poligoni[i++] = {SrediIndeks(poli.a, podaci.brojVrhova),
				SrediIndeks(poli.b, podaci.brojVrhova),
				SrediIndeks(poli.c, podaci.brojVrhova),
				
				SrediIndeks(poli.na, podaci.brojNormala),
				SrediIndeks(poli.nb, podaci.brojNormala),
				SrediIndeks(poli.nc, podaci.brojNormala),
				
				SrediIndeks(poli.ta, podaci.brojUV),
				SrediIndeks(poli.tb, podaci.brojUV),
				SrediIndeks(poli.tc, podaci.brojUV)};
      
      //start += grupa.poligoni.size();		       
    }
}

__host__ void TrokutModel::Dealociraj()
{
  delete[] podaci.vrhovi;
  delete[] podaci.normale;
  delete[] podaci.UV;
  delete[] podaci.poligoni;
  delete[] podaci.transformiraniVrhovi;
  delete[] podaci.transformiraneNormale;
}

__host__ void TrokutModel::Unisti()
{
  Dealociraj();
}

__host__ void TrokutModel::PrimijeniTransformaciju(const Matrica4x4 &mat)
{
  podaci.transformacija =  mat;
}

__host__ void TrokutModel::Realociraj(const TrokutModel &mod2)
{
  std::cout << "TrokutModel Realociraj()\n";
  podaci.brojVrhova = mod2.podaci.brojVrhova;
  podaci.brojNormala = mod2.podaci.brojNormala;
  podaci.brojUV = mod2.podaci.brojUV;
  podaci.brojPoligona = mod2.podaci.brojPoligona;
  podaci.transformacija = mod2.podaci.transformacija;

  Dealociraj();

  podaci.vrhovi = new Vektor3[podaci.brojVrhova];
  podaci.normale = new Vektor3[podaci.brojNormala];
  podaci.UV = new Vektor3[podaci.brojUV];
  podaci.poligoni = new Poligon[podaci.brojPoligona];

  std::memcpy(podaci.vrhovi, mod2.podaci.vrhovi, podaci.brojVrhova * sizeof(Vektor3));
  std::memcpy(podaci.normale, mod2.podaci.normale, podaci.brojNormala * sizeof(Vektor3));
  std::memcpy(podaci.UV, mod2.podaci.UV, podaci.brojUV * sizeof(Vektor3));
  std::memcpy(podaci.poligoni, mod2.podaci.poligoni, podaci.brojPoligona * sizeof(Poligon));
}

__host__ TrokutModel::TrokutModel(const TrokutModel &mod2)
{
  Realociraj(mod2);
}

/*
__device__ __host__ TrokutModel::TrokutModel(TrokutModel &&mod2)
{
  std::cout << "TrokutModel move cons\n";
  Vektor3 *stariVrhovi = podaci.vrhovi;
  Vektor3 *stareNormale = podaci.normale;
  Vektor3 *stariUV = podaci.UV;
  Poligon *stariPoligoni = podaci.poligoni;

  podaci.vrhovi = mod2.podaci.vrhovi;
  podaci.normale = mod2.podaci.normale;
  podaci.UV = mod2.podaci.UV;
  podaci.poligoni = mod2.podaci.poligoni;
  podaci.transformacija = mod2.podaci.transformacija;

  mod2.podaci.vrhovi = stariVrhovi;
  mod2.podaci.normale = stareNormale;
  mod2.podaci.UV = stariUV;
  mod2.podaci.poligoni = stariPoligoni;

  podaci.brojVrhova = mod2.podaci.brojVrhova;
  podaci.brojNormala = mod2.podaci.brojNormala;
  podaci.brojUV = mod2.podaci.brojUV;
  podaci.brojPoligona = mod2.podaci.brojPoligona;
}
*/

__host__ TrokutModel &TrokutModel::operator=(const TrokutModel &mod2)
{
  Realociraj(mod2);

  return *this;
}

__device__ __host__ TrokutModel &TrokutModel::operator=(TrokutModel &&mod2)
{
  std::cout << "TrokutModel move assignment\n";
  podaci.vrhovi = mod2.podaci.vrhovi;
  podaci.normale = mod2.podaci.normale;
  podaci.UV = mod2.podaci.UV;
  podaci.poligoni = mod2.podaci.poligoni;

  podaci.brojVrhova = mod2.podaci.brojVrhova;
  podaci.brojNormala = mod2.podaci.brojNormala;
  podaci.brojUV = mod2.podaci.brojUV;
  podaci.brojPoligona = mod2.podaci.brojPoligona;
  podaci.transformacija = mod2.podaci.transformacija;

  mod2.podaci.vrhovi = nullptr;
  mod2.podaci.normale = nullptr;
  mod2.podaci.UV = nullptr;
  mod2.podaci.poligoni = nullptr;

  return *this;
}

__host__ TrokutModel::~TrokutModel()
{
  //Dealociraj();
}

__host__ RGB TrokutModel::Sjencaj(int presjecenTrokut, const Vektor3 &bariSjeciste, const std::forward_list<Svjetlo> &svjetla)
{
  std::cout << "CPU mesh shader\n";
  Poligon &presjek = podaci.poligoni[presjecenTrokut];
  const Vektor3 &va = podaci.transformiraniVrhovi[presjek.a];
  const Vektor3 &vb = podaci.transformiraniVrhovi[presjek.b];
  const Vektor3 &vc = podaci.transformiraniVrhovi[presjek.c];

  Realan alfa = 1.0f - bariSjeciste[1] - bariSjeciste[2];
  Realan beta = bariSjeciste[1];
  Realan gama = bariSjeciste[2];
  Vektor3 sjeciste(alfa*va + beta*vb + gama*vc);
  
  Vektor3 tn = (alfa*podaci.transformiraneNormale[presjek.na] +
		beta*podaci.transformiraneNormale[presjek.nb] +
		gama*podaci.transformiraneNormale[presjek.nc]).Normiraj();
  
  //Lambert:
  Kanal8 r = 0u, g = 0u, b = 0u;
  for(const Svjetlo &svjetlo : svjetla)
      {
	Vektor3 l = (svjetlo.DajPoziciju().Radijusvektor() - sjeciste).Normiraj();
	Realan m = fmaxf(0.0f, tn * l);
	r += kd[0] * (svjetlo.boja[0] * Kanal8(m));
	g += kd[1] * (svjetlo.boja[1] * Kanal8(m));
	b += kd[2] * (svjetlo.boja[2] * Kanal8(m));
      }

    return RGB{r, g, b};
}


__host__ bool TrokutModel::PresijeciZraku(const Zraka &zraka, Realan &t, int  &presjecenTrokut, Vektor3 &bariSjeciste)
{
  Realan tMin = t;
  bool uspjeh = false;

  bool sijece;
  for(int i = 0; i < podaci.brojPoligona; ++i)
    {
      //backface culling
      Poligon &poligon = podaci.poligoni[i];

      Vektor3 at(podaci.transformiraniVrhovi[poligon.a]);
      Vektor3 bt(podaci.transformiraniVrhovi[poligon.b]);
      Vektor3 ct(podaci.transformiraniVrhovi[poligon.c]);

      const Vektor3 &d = zraka.Smjer();
      const Vektor3 &e = zraka.Ishodiste();

      Vektor3 normala(((bt - at) ^ (ct - at)).Normiraj());
      if(d.Normiraj() * normala > 0)
	continue;
      
      //sijece = PresijeciZrakuTrokut(zraka, t, bariSjeciste);

      Realan eihf = (at[1] - ct[1]) * d[2] - d[1] * (at[2] - ct[2]);
      Realan gfdi = d[0] * (at[2] - ct[2]) - (at[0] - ct[0]) * d[2];
      Realan dheg = (at[0] - ct[0]) * d[1] - (at[1] - ct[1]) * d[0];
      Realan M = (at[0] - bt[0])*eihf + (at[1] - bt[1])*gfdi + (at[2] - bt[2])*dheg;
      
      Realan beta = ((at[0] - e[0])*eihf + (at[1] - e[1])*gfdi + (at[2] - e[2])*dheg) / M;
      Realan gama = ( d[2]*((at[0]-bt[0])*(at[1]-e[1]) - (at[0]-e[0])*(at[1]-bt[1])) + d[1]*((at[0]-e[0])*(at[2]-bt[2]) - (at[0]-bt[0])*(at[2]-e[2])) + d[0]*((at[1]-bt[1])*(at[2]-e[2])-(at[1]-e[1])*(at[2]-bt[2])) ) / M;
      if(beta < 0 || gama < 0 || beta + gama >= 1.0f)
	sijece =  false;
      else
	sijece = true;
      
      t = -( (at[2]-ct[2])*((at[0]-bt[0])*(at[1]-e[1]) - (at[0]-e[0])*(at[1]-bt[1])) + (at[1]-ct[1])*((at[0]-e[0])*(at[2]-bt[2]) - (at[0]-bt[0])*(at[2]-e[2])) + (at[0] - ct[0])*((at[1]-bt[1])*(at[2]-e[2])-(at[1]-e[1])*(at[2]-bt[2])) ) / M;
      
      if(sijece && t < tMin)
	{
	  tMin = t;
	  presjecenTrokut = i;
	  uspjeh = true;
	  bariSjeciste = Vektor3(0.0f, beta, gama);
	}
    }
  
  t = tMin;
  return uspjeh;
}

TrokutModelGPU TrokutModel::PrekopirajNaGPU()
{
  TrokutModelGPU gpuKopija;
  //  gpuKopija.naGPU = true;
  // TrokutModelGPU gpuKopija = GPUCopyable::PrekopirajNaGPU();
  hipMemcpy(&gpuKopija.podaci, &podaci, sizeof(Podaci), hipMemcpyHostToDevice);
  
  hipMalloc(&gpuKopija.podaci.vrhovi, podaci.brojVrhova * sizeof(Vektor3));
  hipMalloc(&gpuKopija.podaci.normale, podaci.brojNormala * sizeof(Vektor3));
  hipMalloc(&gpuKopija.podaci.UV, podaci.brojUV * sizeof(Vektor3));
  hipMalloc(&gpuKopija.podaci.poligoni, podaci.brojPoligona * sizeof(Poligon));

  //ove transformirane vektore će srediti kernel; za sada treba samo rezervirati prostor
  hipMalloc(&gpuKopija.podaci.transformiraniVrhovi, podaci.brojVrhova * sizeof(Vektor3));
  hipMalloc(&gpuKopija.podaci.transformiraneNormale, podaci.brojNormala * sizeof(Vektor3));

  hipMemcpy(gpuKopija.podaci.vrhovi, podaci.vrhovi, podaci.brojVrhova * sizeof(Vektor3), hipMemcpyHostToDevice);
  hipMemcpy(gpuKopija.podaci.normale, podaci.normale, podaci.brojNormala * sizeof(Vektor3), hipMemcpyHostToDevice);
  hipMemcpy(gpuKopija.podaci.UV, podaci.UV, podaci.brojUV * sizeof(Vektor3), hipMemcpyHostToDevice);
  hipMemcpy(gpuKopija.podaci.poligoni, podaci.poligoni, podaci.brojPoligona * sizeof(Poligon), hipMemcpyHostToDevice);

  return std::move(gpuKopija);
  //return gpuKopija;
}
