#pragma once
#include "ObjektGPU.h"

/** Osnovna klasa za trokutaste mesheve.
 *
 *  Ovu klasu nasljeđuju svi koji žele koristiti osnovne podatke i funkcionalnosti vezane za primitivu trokutastih mesheva. U principu su to jedino TrokutModel i TrokutModelGPU.
 */
class TrokutModelBaza : public ObjektGPU
{
 public:
  /** Osnovna struktura za indeksirane poligone.
   *
   *  Postoji opcija uključivanja i teksturnih koordinata i vršnih normala.
   */
  struct Poligon
  {
    int a, b, c; //indeksi vrhova
    int na, nb, nc; //indeksi normala; -1 znači da normala nije specificirana
    int ta, tb, tc; //uv indeksi; -1 znači da UV nisu specificirani
  };

  /** Struktura Osnovnih podataka o meshu.
   */
  struct Podaci
  {
    size_t brojVrhova = 0;
    size_t brojNormala = 0;
    size_t brojUV = 0;
    size_t brojPoligona = 0;
    Vektor3 *vrhovi = nullptr;
    Vektor3 *normale = nullptr;
    Vektor3 *UV = nullptr;
    //Materijal materijal; //TODO: materijali
    Poligon *poligoni = nullptr;
    
    Vektor3 *transformiraniVrhovi = nullptr;
    Vektor3 *transformiraneNormale = nullptr;
    Matrica4x4 transformacija;
  };
  /*
  __host__ TrokutModelBaza(const TrokutModel &orig);
  __host__ TrokutModelBaza() = default;
  __host__ TrokutModelBaza(const ModelOBJ::Objekt &obj);
  __host__ TrokutModelBaza(const TrokutModelBaza&);
  __device__ __host__ TrokutModelBaza(TrokutModelBaza&&);
  __host__ TrokutModelBaza &operator=(const TrokutModelBaza&);
  __device__ __host__ TrokutModelBaza &operator=(TrokutModelBaza&&);
  __host__ ~TrokutModelBaza();
  */
  
 protected:
  /// Podaci podaci;
  /** Difuzna boja površine mesha.
   *
   *  U nedostatku cijelog sustava za materijale, ovo služi za postavljanje različitih izgleda mesheva.
   */
  RGB kd = {1.0f, 0.5f, 0.5f};

 public:
   Podaci podaci;
   
  __device__ __host__ int DajBrojNitiZaTransform() const override {return podaci.brojVrhova;}

};
