#include "TrokutModelGPU.h"

/*
__host__ std::unique_ptr<char[]> TrokutModelGPU::DajPodatke()
{
  auto data = std::unique_ptr<char[]>{new char[sizeof(Podaci)]};
  std::memcpy(data.get(), &podaci, sizeof(Podaci));
  return data;
}

__host__ void TrokutModelGPU::ZapisiPodatke(std::unique_ptr<char[]> data)
{
  std::memcpy(&podaci, data.get(), sizeof(Podaci));
}
*/

__device__ __host__ TrokutModelGPU::TrokutModelGPU(const TrokutModelBaza::Podaci &novi)
{
  podaci = novi;
  //std::cout << "Prekopirani podaci ok" << std::endl;
} //TODO: neka ovako radi interface GPUResident: ima ovaj konstruktor i 'podaci' bazu svih karakterističnih podataka za GPU objekt (minus ostali privatni implementacijski detalji koji ostaju izvan)

__device__ __host__ TrokutModelGPU::~TrokutModelGPU()
{

}

__host__ void TrokutModelGPU::Unisti()
{
  if(podaci.vrhovi)
    hipFree(podaci.vrhovi);
  if(podaci.normale)
    hipFree(podaci.normale);
  if(podaci.UV)
    hipFree(podaci.UV);
  if(podaci.poligoni)
    hipFree(podaci.poligoni);

  hipFree(podaci.transformiraniVrhovi);
  hipFree(podaci.transformiraneNormale);
}

__device__ void TrokutModelGPU::Transformiraj(int offset)
{
  int i = offset;
  podaci.transformiraniVrhovi[i] = (podaci.transformacija * podaci.vrhovi[i].DajTocku()).Radijusvektor();
  if(i < podaci.brojNormala)
    podaci.transformiraneNormale[i] = podaci.transformacija.Linearna() * podaci.normale[i];
}

__device__ bool TrokutModelGPU::PresijeciZraku(const Zraka &zraka, Realan &t, int &presjecenTrokut, Vektor3 &bariSjeciste)
{
  ///Realan tMin = BESKONACNO;
  Realan tMin = t;
  bool uspjeh = false;

  bool sijece;
  for(int i = 0; i < podaci.brojPoligona; ++i)
    {
      //backface culling
      Poligon &poligon = podaci.poligoni[i];

      Vektor3 at(podaci.transformiraniVrhovi[poligon.a]);
      Vektor3 bt(podaci.transformiraniVrhovi[poligon.b]);
      Vektor3 ct(podaci.transformiraniVrhovi[poligon.c]);

      const Vektor3 &d = zraka.Smjer();
      const Vektor3 &e = zraka.Ishodiste();

      Vektor3 normala(((bt - at) ^ (ct - at)).Normiraj());
      if(d.Normiraj() * normala > 0)
	continue;
      
      Realan eihf = (at[1] - ct[1]) * d[2] - d[1] * (at[2] - ct[2]);
      Realan gfdi = d[0] * (at[2] - ct[2]) - (at[0] - ct[0]) * d[2];
      Realan dheg = (at[0] - ct[0]) * d[1] - (at[1] - ct[1]) * d[0];
      Realan M = (at[0] - bt[0])*eihf + (at[1] - bt[1])*gfdi + (at[2] - bt[2])*dheg;
      
      Realan beta = ((at[0] - e[0])*eihf + (at[1] - e[1])*gfdi + (at[2] - e[2])*dheg) / M;
      Realan gama = ( d[2]*((at[0]-bt[0])*(at[1]-e[1]) - (at[0]-e[0])*(at[1]-bt[1])) + d[1]*((at[0]-e[0])*(at[2]-bt[2]) - (at[0]-bt[0])*(at[2]-e[2])) + d[0]*((at[1]-bt[1])*(at[2]-e[2])-(at[1]-e[1])*(at[2]-bt[2])) ) / M;
      if(beta < 0 || gama < 0 || beta + gama >= 1.0f)
	sijece =  false;
      else
	sijece = true;
      
      t = -( (at[2]-ct[2])*((at[0]-bt[0])*(at[1]-e[1]) - (at[0]-e[0])*(at[1]-bt[1])) + (at[1]-ct[1])*((at[0]-e[0])*(at[2]-bt[2]) - (at[0]-bt[0])*(at[2]-e[2])) + (at[0] - ct[0])*((at[1]-bt[1])*(at[2]-e[2])-(at[1]-e[1])*(at[2]-bt[2])) ) / M;
      
      if(sijece && t < tMin)
	{
	  tMin = t;
	  presjecenTrokut = i;
	  uspjeh = true;
	  bariSjeciste = Vektor3(0.0f, beta, gama);
	}
    }
  
  t = tMin;
  return uspjeh;
}

__device__ RGB TrokutModelGPU::Sjencaj(int presjecenTrokut, const Vektor3 &bariSjeciste, const Svjetlo svjetla[], size_t n)
{
  Poligon &presjek = podaci.poligoni[presjecenTrokut];
  const Vektor3 &va = podaci.transformiraniVrhovi[presjek.a];
  const Vektor3 &vb = podaci.transformiraniVrhovi[presjek.b];
  const Vektor3 &vc = podaci.transformiraniVrhovi[presjek.c];

  Realan alfa = 1.0f - bariSjeciste[1] - bariSjeciste[2];
  Realan beta = bariSjeciste[1];
  Realan gama = bariSjeciste[2];
  Vektor3 sjeciste(alfa*va + beta*vb + gama*vc);
  
  Vektor3 tn = (alfa*podaci.transformiraneNormale[presjek.na] +
		beta*podaci.transformiraneNormale[presjek.nb] +
		gama*podaci.transformiraneNormale[presjek.nc]).Normiraj();
  

  //  return SjencajTrokut(tn, sjeciste, svjetla, n);
  //Lambert:
  Kanal8 r = 0u, g = 0u, b = 0u;
    for(int i = 0; i < n; ++i)
      {
	const Svjetlo &svjetlo = svjetla[i];
	Vektor3 l = (svjetlo.DajPoziciju().Radijusvektor() - sjeciste).Normiraj();
	Realan m = fmaxf(0.0f, tn * l);
	r += kd[0] * (svjetlo.boja[0] * Kanal8(m));
	g += kd[1] * (svjetlo.boja[1] * Kanal8(m));
	b += kd[2] * (svjetlo.boja[2] * Kanal8(m));
      }

    return RGB{r, g, b};
  
}
