#pragma once
#include "TrokutModel.h"
#include "SjeceTrokut.h"
#include "SjencaTrokut.h"
#include "GPUResident.h"
#include "TrokutModelBaza.h"
#include <memory>

class TrokutModel;
class ScenaGPU;

/** Tip indeksiranog trokutastog mesha na GPU-u.
 *
 *  Trebao bi se konstruirati kopiranjem CPU TrokutModela radi smještanja u odgovarajuću scenu.
 */
class TrokutModelGPU : public TrokutModelBaza, public GPUResident<TrokutModelBaza::Podaci>
{
 public:
  friend TrokutModel;
  friend ScenaGPU;
    
  __host__ TrokutModelGPU() = default;
  __host__ __device__ TrokutModelGPU(const TrokutModelBaza::Podaci&);
  __device__ __host__ TrokutModelGPU(const TrokutModel &mod){}
  __host__ TrokutModelGPU(const TrokutModelGPU&) = default;
  __host__ TrokutModelGPU &operator=(const TrokutModelGPU&) = default;
  __host__ TrokutModelGPU(TrokutModelGPU&&) = default;
  __device__ __host__ TrokutModelGPU &operator=(TrokutModelGPU&&) = default;
  __device__ __host__ ~TrokutModelGPU();

  __host__ Podaci DajPodatke() override {return podaci;}
  

  /*
  __host__ std::unique_ptr<char[]> DajPodatke() override;
  __host__ void ZapisiPodatke(std::unique_ptr<char[]> data) override;  */

  __device__ void Transformiraj(int offset) override;

  __device__ bool PresijeciZraku(const Zraka &zraka, Realan &t, int &presjecenTrokut, Vektor3 &bariSjeciste);
  
  __device__ RGB Sjencaj(int presjecenTrokut, const Vektor3 &sjeciste, const Svjetlo[], size_t);

 private:
 protected:
  __host__ void UpisiPodatke(const Podaci &novi) override {podaci = novi;}
  __host__ void Unisti() override;
};
