#include "Vektor4.h"
#include "Vektor3.h"
#include <iostream>

__device__ __host__ Vektor3::Vektor3() : x(0), y(0), z(0) {}

__device__ __host__ Vektor3::Vektor3(Realan x, Realan y, Realan z)
{
  this->x = x;
  this->y = y;
  this->z = z;
}

__device__ __host__ Vektor3 Vektor3::operator+(const Vektor3 &v) const
{
  Vektor3 tmp = v;
  Realan a = x + tmp[0];
  Realan b = y + tmp[1];
  Realan c = z + tmp[2];

  return Vektor3(a, b, c);
}

__device__ __host__ Vektor3 &Vektor3::operator+=(const Vektor3 &v)
{
  *this = *this + v;
  return *this;
}

__device__ __host__ Vektor3 Vektor3::operator-(const Vektor3 &v) const
{
  Vektor3 tmp = v;
  Realan a = x - tmp[0];
  Realan b = y - tmp[1];
  Realan c = z - tmp[2];

  return Vektor3(a, b, c);
}

__device__ __host__ Vektor3 Vektor3::operator-() const
{
  return Vektor3(-x, -y, -z);
}

__device__ __host__ Vektor3 operator-(const Vektor3 &v)
{
  Vektor3 tmp = v;
  return Vektor3(-tmp[0], -tmp[1], -tmp[2]);
}

//void Vektor3::operator-()
//{
//  x = -x;
//  y = -y;
//  z = -z;
//}

__device__ __host__ Vektor3 Vektor3::operator^(const Vektor3 &v) const
{
  Vektor3 tmp = v;
#ifdef FP_FAST_FMA
  Realan a = fma (y, tmp[2], -z * tmp[1]);
  Realan b = fma (z, tmp[0], -x * tmp[2]);
  Realan c = fma (x, tmp[1], -y * tmp[0]);
#else
  Realan a = y * tmp[2] - z * tmp[1];
  Realan b = z * tmp[0] - x * tmp[2];
  Realan c = x * tmp[1] - y * tmp[0];
#endif

  return Vektor3(a, b, c);
}

__device__ __host__ Realan Vektor3::operator*(const Vektor3 &v) const
{
  Vektor3 tmp = v;
#ifdef FP_FAST_FMA
  return fma (x, tmp[0], fma (y, tmp[1], z * tmp[2]));
#else
  return (x * tmp[0] + y * tmp[1] + z * tmp[2]);
#endif
}

__device__ __host__ Vektor3 operator*(const Matrica3x3 &mat, const Vektor3 &vek)
{
  Vektor3 tmp = vek;
#ifdef FP_FAST_FMA
  Realan a = fma (mat.m[0][0], tmp[0], fma (mat.m[0][1], tmp[1], mat.m[0][2] * tmp[2]));
  Realan b = fma (mat.m[1][0], tmp[0], fma (mat.m[1][1], tmp[1], mat.m[1][2] * tmp[2]));
  Realan c = fma (mat.m[2][0], tmp[0], fma (mat.m[2][1], tmp[1], mat.m[2][2] * tmp[2]));
#else
  Realan a = mat.m[0][0] * tmp[0] + mat.m[0][1] * tmp[1] + mat.m[0][2] * tmp[2];
  Realan b = mat.m[1][0] * tmp[0] + mat.m[1][1] * tmp[1] + mat.m[1][2] * tmp[2];
  Realan c = mat.m[2][0] * tmp[0] + mat.m[2][1] * tmp[1] + mat.m[2][2] * tmp[2];
#endif

  return Vektor3(a, b, c);
}

__device__ __host__ Vektor3 operator*(const Vektor3 &vek, Realan skalar)
{
  Vektor3 tmp = vek;

  Realan a = tmp[0] * skalar;
  Realan b = tmp[1] * skalar;
  Realan c = tmp[2] * skalar;

  return Vektor3(a, b, c);
}

__device__ __host__ Vektor3 operator*(Realan skalar, const Vektor3 &vek)
{
  return operator*(vek, skalar);
}

__device__ __host__ Realan Vektor3::operator[](int indeks) const
{
  switch (indeks)
    {
    case 0:
      return x;

    case 1:
      return y;

    case 2:
      return z;
    }
}

__device__ __host__ Realan Vektor3::EuklidskaNorma() const
{
  //return hypot(x, hypot(y,z));
#ifdef FP_FAST_FMA
  return sqrt(fma (x, x, fma (y, y, z * z)));
#else
  return sqrt (x * x + y * y + z * z);
#endif
}

__device__ __host__ Vektor3 Vektor3::Normiraj() const
{
  return *this * (1 / EuklidskaNorma());
}

__device__ __host__ Realan Vektor3::MaxNorma() const
{
  return fmax(fabs(x), fmax(fabs(y), fabs(z)));
}

__device__ __host__ Vektor4 Vektor3::DajTocku() const
{
  return Vektor4(x, y, z, 1);
}

__device__ __host__ RGB Vektor3::DajRGB()
{
  unsigned int r, g, b;
  if(x > 255)
    x = 255;
  else if(x < 0)
    x = 0;

  if(y > 255)
    y = 255;
  else if(y < 0)
    y = 0;

  if(z > 255)
    z = 255;
  else if(z < 0)
    z = 0;

  r = static_cast<unsigned int>(x);
  g = static_cast<unsigned int>(y);
  b = static_cast<unsigned int>(z);
  return RGB{r, g, b};
}

__device__ __host__ void Vektor3::DajKomponente(Realan &x, Realan &y, Realan &z) const
{
  x = this->x;
  y = this->y;
  z = this->z;
}
