#pragma once
//#include "Vektor4.h"
#include "MatBaza.h"
#include "RGB.h"
class Vektor4;

/** Tip trodimenzionalnog vektora.
 */
class Vektor3
{
 private:
  /** Koordinate.
   */
  Realan x, y, z;//koordinate

 public:
  /** Množenje vektora sa skalarom.
   *  \param[in] skalar Realan skalar.
   *  \param[in] vektor Vektorski operand.
   *  \return Produktni vektor.
   */
  __device__ __host__ friend Vektor3 operator*(Realan skalar,const Vektor3 &vektor);
  __device__ __host__ friend Vektor3 operator*(const Vektor3 &vektor,Realan skalar);
  //__device__ __host__ friend Vektor3 operator-(const Vektor3 &);
  
  /** Konstruira nulvektor.
   */
  __device__ __host__  Vektor3();
  /** Konstruira 3-vektor iz 3 dana elementa.
   *  \param[in] a,b,c Dani elementi.
   */
  __device__ __host__  Vektor3(Realan a,Realan b,Realan c);
  __device__ __host__ ~Vektor3() = default;

  __device__ __host__ Vektor3(const Vektor3 &) = default;
  __device__ __host__ Vektor3 & operator=(const Vektor3 &) = default;
  __device__ __host__ Vektor3(Vektor3 &&) = default;
  __device__ __host__ Vektor3 & operator=(Vektor3 &&) = default;

  /** Zbraja dva 3-vektora.
   *  \param[in] drugi  Desni vektorski operand.
   *  \return Novi 3-vektor sume.
   */
  __device__ __host__ Vektor3 operator+(const Vektor3 &drugi) const;
  /** Oduzima dva 3-vektora.
   *  \param[in] drugi  Desni vektorski operand.
   *  \return Novi 3-vektor razlike.
   */
  __device__ __host__ Vektor3 operator-(const Vektor3 &drugi) const;
  /** Negira 3-vektor.
   *  \return Novi negirani 3-vektor.
   */
  __device__ __host__ Vektor3 operator-() const;
  /** Kombinirani operator zbrajanja i pridruživanja 3-vektora.
   *  \param[in] drugi  Desni vektorski operand.
   *  \return Referenca na ovaj 3-vektor.
   */
  __device__ __host__ Vektor3 &operator+=(const Vektor3 &drugi);

  /** Računa vektorski produkt ovog i drugog danog 3-vektora.
   *  \param[in] drugi  Desni operand vektorskog produkta.
   *  \return Vektorski produkt ovog i danog 3-vektora.
   */
  __device__ __host__ Vektor3 operator^(const Vektor3 &drugi) const;//vektorski produkt
  /** Računa skalarni produkt ovog i drugog danog 3-vektora.
   *  \param[in] drugi  Desni operand skalarnog produkta.
   *  \return Skalar koji je rezultat skalarnog produkta.
   */
  __device__ __host__ Realan operator*(const Vektor3 &drugi) const;//skalarni produkt

  /** Vraća \f$i\f$-ti element vektora.
   *  \param[in] i Indeks
   */
  __device__ __host__ Realan operator[](int i) const;

  /** Računa standardnu euklidsku normu vektora.
   */
  __device__ __host__ Realan EuklidskaNorma() const;
  /** Vraća normiran vektor.
   */
  __device__ __host__ Vektor3 Normiraj() const;
  __device__ __host__ Realan MaxNorma() const;

  /** Vraća točku koja odgovara radijusvektoru.
   *  Vraća 4-vektor s \f$1.0\f$ u posljednjoj komponenti, gdje su prethodne komponente identične onima iz trenutnog 3-vektora.
   */
  __device__ __host__ Vektor4 DajTocku() const;
  __device__ __host__ RGB DajRGB();

  /** Vraća sve komponente vektora po referenci.
   *  \param[out] x,y,z Komponente.
   */
  __device__ __host__ void DajKomponente(Realan &x,Realan &y,Realan &z) const;
};
