#include "Vektor4.h"
#include <iostream>
//#include <cmath>

__device__ __host__ Vektor4::Vektor4() : x(0), y(0), z(0), w(0) {}

__device__ __host__ Vektor4::Vektor4(Realan x, Realan y, Realan z, Realan w)
{
  this->x = x;
  this->y = y;
  this->z = z;
  this->w = w;
}

__device__ __host__ Vektor4 Vektor4::operator+(const Vektor4 &v) const
{
  Vektor4 tmp = v;
  Realan a = x + tmp[0];
  Realan b = y + tmp[1];
  Realan c = z + tmp[2];
  Realan d = w + tmp[3];

  return Vektor4(a, b, c, d);
}

__device__ __host__ Vektor4 Vektor4::operator-(const Vektor4 &v) const
{
  Vektor4 tmp = v;
  Realan a = x - tmp[0];
  Realan b = y - tmp[1];
  Realan c = z - tmp[2];
  Realan d = w - tmp[3];

  return Vektor4(a, b, c, d);
}

__device__ __host__ Vektor4 operator-(const Vektor4 &v)
{
  Vektor4 tmp = v;
  return Vektor4(-tmp[0], -tmp[1], -tmp[2], -tmp[3]);
}

__device__ __host__ Vektor4 Vektor4::operator-() const
{
  return Vektor4(-x, -y, -z, -w);
}

__device__ __host__ Realan Vektor4::operator*(const Vektor4 &v)
{
  Vektor4 tmp = v;
#ifdef FP_FAST_FMA
  return fma (x, tmp[0], fma (y, tmp[1], fma (z, tmp[2], w * tmp[3])));
#else
  return (x * tmp[0] + y * tmp[1] + z * tmp[2] + w * tmp[3]);
#endif
}

__device__ __host__ Vektor4 operator*(const Matrica4x4 &mat, const Vektor4 &vek)
{
  Vektor4 tmp = vek;
#ifdef FP_FAST_FMA
  Realan tmp1 = fma (mat.m[0][2], tmp[2], mat.m[0][3] * tmp[3]);
  Realan tmp2 = fma (mat.m[0][1], tmp[1], tmp1);
  Realan a = fma (mat.m[0][0], tmp[0], tmp2);

  tmp1 = fma (mat.m[1][2], tmp[2], mat.m[1][3] * tmp[3]);
  tmp2 = fma (mat.m[1][1], tmp[1], tmp1);
  Realan b = fma (mat.m[1][0], tmp[0], tmp2);

  tmp1 = fma (mat.m[2][2], tmp[2], mat.m[2][3] * tmp[3]);
  tmp2 = fma (mat.m[2][1], tmp[1], tmp1);
  Realan c = fma (mat.m[2][0], tmp[0], tmp2);

  tmp1 = fma (mat.m[3][2], tmp[2], mat.m[3][3] * tmp[3]);
  tmp2 = fma (mat.m[3][1], tmp[1], tmp1);
  Realan d = fma (mat.m[3][0], tmp[0], tmp2);
#else
  Realan a = mat.m[0][0] * tmp[0] + mat.m[0][1] * tmp[1] + mat.m[0][2] * tmp[2] + mat.m[0][3] * tmp[3];
  Realan b = mat.m[1][0] * tmp[0] + mat.m[1][1] * tmp[1] + mat.m[1][2] * tmp[2] + mat.m[1][3] * tmp[3];
  Realan c = mat.m[2][0] * tmp[0] + mat.m[2][1] * tmp[1] + mat.m[2][2] * tmp[2] + mat.m[2][3] * tmp[3];
  Realan d = mat.m[3][0] * tmp[0] + mat.m[3][1] * tmp[1] + mat.m[3][2] * tmp[2] + mat.m[3][3] * tmp[3];
  #endif

  return Vektor4(a, b, c, d);
}

__device__ __host__ Vektor4 operator*(const Vektor4 &vek, Realan skalar)
{
  Vektor4 tmp = vek;

  Realan a = tmp[0] * skalar;
  Realan b = tmp[1] * skalar;
  Realan c = tmp[2] * skalar;
  Realan d = tmp[3] * skalar;

  return Vektor4(a, b, c, d);
}

__device__ __host__ Vektor4 operator*(Realan skalar, const Vektor4 &vek)
{
  return operator*(vek, skalar);
}

__device__ __host__ Realan Vektor4::operator[](int indeks)
{
  switch (indeks)
    {
    case 0:
      return x;

    case 1:
      return y;

    case 2:
      return z;

    case 3:
      return w;
    }
}

__device__ __host__ Vektor3 Vektor4::PerspektivnoDijeljenje() const
{
  if (w == 1.0)
    return Vektor3(x, y, z);
  //else
  Realan a = x / w;
  Realan b = y / w;
  Realan c = z / w;

  return Vektor3(a, b, c);//bolje Vektor4(a, b, c, 1);?
}

__device__ __host__ Realan Vektor4::EuklidskaNorma() const
{
  //return hypot(x, hypot(y,z));
#ifdef FP_FAST_FMA
  return sqrt (fma (x, x, fma (y, y, fma (z, z, w * w))));
#else
  return sqrt (x * x + y * y + z * z + w * w);
#endif
}

__device__ __host__ Vektor3 Vektor4::Radijusvektor() const
{
  return Vektor3(x, y, z);
}
