#pragma once
#include "MatBaza.h"
#include "Matrica4x4.h"
#include "Vektor3.h"

/** Tip četverodimenzionalnog vektora.
 */
class Vektor4
{
 private:
  /** Koordinate.
   */
  Realan x, y, z, w;//koordinate

 public:
  /** Množenje vektora sa skalarom.
   *  \param[in] skalar Realan skalar.
   *  \param[in] vektor Vektorski operand.
   *  \return Produktni vektor.
   */
  friend __device__ __host__ Vektor4 operator*(Realan,const Vektor4 &);
  friend __device__ __host__ Vektor4 operator*(const Vektor4 &,Realan);
  /** Negira dani 4-vektor.
   *  \param[in] vek Dani vektor.
   *  \return Novi negirani 4-vektor.
   */
  friend __device__ __host__ Vektor4 operator-(const Vektor4 &vek);

  /** Konstruira nulvektor.
   */
  __device__ __host__ Vektor4();
  /** Konstruira 4-vektor iz 4 dana elementa.
   *  \param[in] a,b,c,d Dani elementi.
   */
  __device__ __host__ Vektor4(Realan a,Realan b,Realan c,Realan d);
  __device__ __host__ ~Vektor4() = default;

  __device__ __host__ Vektor4(const Vektor4 &) = default;
  __device__ __host__ Vektor4 & operator=(const Vektor4 &) = default;
  __device__ __host__ Vektor4(Vektor4 &&) = default;
  __device__ __host__ Vektor4 & operator=(Vektor4 &&) = default;

  /** Zbraja dva 4-vektora.
   *  \param[in] drugi  Desni vektorski operand.
   *  \return Novi 4-vektor sume.
   */
  __device__ __host__ Vektor4 operator+(const Vektor4 &drugi) const;
  /** Oduzima dva 4-vektora.
   *  \param[in] drugi  Desni vektorski operand.
   *  \return Novi 4-vektor razlike.
   */
  __device__ __host__ Vektor4 operator-(const Vektor4 &drugi) const;
  __device__ __host__ Vektor4 operator-() const;

  /** Računa skalarni produkt ovog i drugog danog 4-vektora.
   *  \param[in] drugi  Desni operand skalarnog produkta.
   *  \return Skalar koji je rezultat skalarnog produkta.
   */
  __device__ __host__ Realan operator*(const Vektor4 &drugi);//skalarni produkt

  /** Vraća \f$i\f$-ti element vektora.
   *  \param[in] i Indeks
   */
  __device__ __host__ Realan operator[](int i);

  /** Projicira točku u 3-vektor.
   *  Računa se vektor \f$(x/w,y/w,z/w,1)\f$ i vraća samo prve tri navedene koordinate.
   */
  __device__ __host__ Vektor3 PerspektivnoDijeljenje() const;

  __device__ __host__ Realan EuklidskaNorma() const;
  __device__ __host__ Realan MaxNorma() const;

  /** Vraća odgovarajući radijusvektor.
   *  \return 3-vektor \f$(x,y,z)\f$.
   */
  __device__ __host__ Vektor3 Radijusvektor() const;
};
